
/*
 * This represents one entry in a distributed enum.
 */

public class DataDistributedEnumEntry extends DataComponent
{
	/* Class variables */
  public final static String   ID           = "ID";
  public final static String   EQUALS       = "=";
  public final static String   EQUALS_VALUE = "VALUE";

	/* Instance Variables */
  protected String           id;
  protected String           positiveIntegerEqualsString;
  protected DataExpression   generalEqualsExpression;
  protected boolean          usingPositiveIntegerEqualsString;


	/* Instance Methods */
  public DataDistributedEnumEntry()
  {
    id                               = null;
    positiveIntegerEqualsString      = null;
    generalEqualsExpression          = null;
    usingPositiveIntegerEqualsString = true;
  }

  public String getId() { return id; }
  public void   setId ( String theId ) { id = theId; }

  public String          getPositiveIntegerEqualsString()
				   { return positiveIntegerEqualsString; }

  public DataExpression  getGeneralEqualsExpression()
				   { return generalEqualsExpression; }

  public boolean         getUsingPositiveIntegerEqualsString()
				   { return usingPositiveIntegerEqualsString; }


  public void    setPositiveIntegerEqualsString ( String theEqualsString )
  {
    positiveIntegerEqualsString      = theEqualsString;
    usingPositiveIntegerEqualsString = true;
  }

  public void    setGeneralEqualsExpression ( DataExpression theExpression )
  {
    generalEqualsExpression          = theExpression;
    usingPositiveIntegerEqualsString = false;
  }


  public boolean hasEquals()
  {
    if ( getUsingPositiveIntegerEqualsString() == true )
      return getPositiveIntegerEqualsString()  != null;
    else
      return getGeneralEqualsExpression()      != null;
  }




  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );
    return theTDLParser . parseDataDistributedEnumEntry ( null );
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    if ( /*NOT*/ ! ( theObjectToCopy instanceof DataDistributedStructOrEnum ) )
    {
      System.err.println ( "[DataDistributedStructOrEnum:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataDistributedEnumEntry  theEnumEntryToCopy
			        = (DataDistributedEnumEntry) theObjectToCopy;

    id = theEnumEntryToCopy.id;

    positiveIntegerEqualsString
      = theEnumEntryToCopy.positiveIntegerEqualsString;

    if (   ( theCanDestroyObject                        == true )
	|| ( theEnumEntryToCopy.generalEqualsExpression == null ) )
      generalEqualsExpression = theEnumEntryToCopy.generalEqualsExpression;
    else
      generalEqualsExpression
	= ((DataExpression)
	   (theEnumEntryToCopy.generalEqualsExpression . copyComponent()) );

    usingPositiveIntegerEqualsString
      = theEnumEntryToCopy.usingPositiveIntegerEqualsString;

    return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

    initializeGenerateSubcomponentIndex();


	/* Write out ID */
    generateSubcomponents ( DataDistributedEnumEntry.ID, theOutputDestination,
			    theObjectSubsetToGenerate, false );
    theOutputDestination . write ( getId() );

	/* Deal with equals */
    if ( hasEquals() == true )
    {
      generateSubcomponents ( DataDistributedEnumEntry.EQUALS,
			      theOutputDestination,
			      theObjectSubsetToGenerate, false );
      theOutputDestination . write ( DataDistributedEnumEntry.EQUALS );



      generateSubcomponents ( DataDistributedEnumEntry.EQUALS_VALUE,
			      theOutputDestination,
			      theObjectSubsetToGenerate, false );

      if ( getUsingPositiveIntegerEqualsString() == true )
      {
	theOutputDestination . write ( getPositiveIntegerEqualsString() );
      }
      else
      {
	getGeneralEqualsExpression() . generate ( theOutputDestination,
						  theObjectSubsetToGenerate );
      }
    }

	/* Deal with close-brace, semicolon, and other leftovers. */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }

}
