
/*
 * This represents one of:
 *      1) DISTRIBUTED struct [ID] {  (decls;)+       } ;
 *      2) DISTRIBUTED enum   [ID] {  (name[=value])+ } ;
 */

public class DataDistributedStructOrEnum extends DataComponent
{
	/* Class Variables */
  public final static String   DISTRIBUTED    = "DISTRIBUTED";
  public final static String   FIRST_ID_INDEX = "FirstId";

	/* Instance Variables */
  protected String   firstId;
  protected boolean  isStruct;

	/* Instance Methods */
  public DataDistributedStructOrEnum()
  {
    firstId  = null;
    isStruct = true;
  }

  public String getFirstId () { return firstId; }
  public void   setFirstId ( String theFirstId ) { firstId = theFirstId; }


  public boolean getIsStruct() { return isStruct;  }
  public void    setIsStruct() { isStruct = true;  }
  public void    setIsEnum()   { isStruct = false; }

  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );
    return theTDLParser . parseDistributedStructOrEnum();
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    if ( /*NOT*/ ! ( theObjectToCopy instanceof DataDistributedStructOrEnum ) )
    {
      System.err.println ( "[DataDistributedStructOrEnum:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataDistributedStructOrEnum  theStructOrEnumToCopy
			       = (DataDistributedStructOrEnum) theObjectToCopy;

    firstId  = theStructOrEnumToCopy.firstId;
    isStruct = theStructOrEnumToCopy.isStruct;

    return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

    initializeGenerateSubcomponentIndex();


	/* Write out DISTRIBUTED */
    generateSubcomponents ( DataDistributedStructOrEnum.DISTRIBUTED,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );
    if ( hasIndex ( DataDistributedTypedef.DISTRIBUTED ) )
    {
      if ( theObjectSubsetToGenerate != ENTIRE_OBJECT )
	theOutputDestination . write ( "/*" );
      theOutputDestination . write ( DataDistributedTypedef.DISTRIBUTED );
      if ( theObjectSubsetToGenerate != ENTIRE_OBJECT )
	theOutputDestination . write ( "*/" );
    }


    generateSubcomponents ( DataDistributedStructOrEnum.FIRST_ID_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );
    if ( getFirstId() != null )
      theOutputDestination . write ( getFirstId() );
    

	/* Deal with close-brace, semicolon, and other leftovers. */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }

}
