
import java.util.Vector;

/*
 * This represents:
 *      DISTRIBUTED typedef {distributed-type} ID (, [const][*][const] ID)*
 *
 * Uses DataTaskArgument to hold the distributed-type.
 */

public class DataDistributedTypedef extends DataComponent
{
	/* Class Variables */
  public final static String   DISTRIBUTED         = "DISTRIBUTED";
  public final static String   TASK_ARGUMENT_INDEX = "taskArgument";
  public final static String   ID_INDEX            = "I";

	/* Instance Variables */
  protected DataTaskArgument dataTaskArgument;
  protected Vector           typedefIdVector;

	/* Instance Methods */
  public DataDistributedTypedef()
  {
    typedefIdVector = new Vector();
  }


  protected Vector getTypedefIdVector() { return typedefIdVector; }

  public void addTypedefIdentifier ( String theIdentifier )
  {
    getTypedefIdVector() . addElement ( theIdentifier );
  }

  public DataTaskArgument getDataTaskArgument() { return dataTaskArgument; }
  public void   setDataTaskArgument( DataTaskArgument theDataTaskArgument )
				 { dataTaskArgument = theDataTaskArgument; }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );
    return theTDLParser . parseDistributedTypedef();
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    if ( /*NOT*/ ! ( theObjectToCopy instanceof DataDistributedTypedef ) )
    {
      System.err.println ( "[DataDistributedTypedef:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataDistributedTypedef  theDistributedTypedefToCopy
			       = (DataDistributedTypedef) theObjectToCopy;

    if ( theCanDestroyObject == true )
    {
      dataTaskArgument = theDistributedTypedefToCopy.dataTaskArgument;
      typedefIdVector  = theDistributedTypedefToCopy.typedefIdVector;
    }
    else
    {
      dataTaskArgument = ((DataTaskArgument)
			  (theDistributedTypedefToCopy . dataTaskArgument
			                               . copyComponent() ) );
      typedefIdVector = (Vector) theDistributedTypedefToCopy . typedefIdVector
	                                                     . clone();
    }

    return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

    initializeGenerateSubcomponentIndex();


	/* Write out DISTRIBUTED */
    generateSubcomponents ( DataDistributedTypedef.DISTRIBUTED,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );
    if ( hasIndex ( DataDistributedTypedef.DISTRIBUTED ) )
    {
      if ( theObjectSubsetToGenerate != ENTIRE_OBJECT )
	theOutputDestination . write ( "/*" );
      theOutputDestination . write ( DataDistributedTypedef.DISTRIBUTED );
      if ( theObjectSubsetToGenerate != ENTIRE_OBJECT )
	theOutputDestination . write ( "*/" );
    }


	/* Write out DataTaskArgument */
    generateSubcomponents ( DataDistributedTypedef.TASK_ARGUMENT_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );
    getDataTaskArgument() . generate ( theOutputDestination,
				       theObjectSubsetToGenerate );

	/* Write out each Identifier */
    for ( int i = 0;  i < getTypedefIdVector() . size();  i++ )
    {
      generateSubcomponents ( (DataDistributedTypedef.ID_INDEX + i),
			      theOutputDestination,
			      theObjectSubsetToGenerate, false );
      theOutputDestination . write( (String)
				    (getTypedefIdVector() . elementAt ( i )) );
    }

	/* Deal with close-brace, semicolon, and other leftovers. */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }

}
