
import netscape.util.*;
import netscape.application.Color;

/*
 * This represents a Task being spawned off.
 */

public class DataSpawnTask extends DataStatement
                        implements DataValidateCode, DataConstrainedObject
{
	/* Class variables */
  public final static String   SPAWN            = "SPAWN";
  public final static String   TASK_NAME_INDEX  = "taskNameIndex";
  public final static String   OPEN_PAREN       = "(";
  public final static String   CLOSE_PAREN      = ")";
  public final static String   COMMA            = ",";
  public final static String   WITH             = "WITH";
  public final static String   SEMICOLON        = ";";

  public       static Color    SPAWN_COLOR      = Color.red;
  public final static int      TASK_ONLY        = 3;
  public final static int      CONSTRAINTS_ONLY = 2;

  protected    static int      ARBITRARY_SUBTASK_INDEX = 0;

	/* Sometimes we need to generate() a wait constraint... */
  protected    static DataConstraint   WAIT_CONSTRAINT
				  = new DataConstraint ( DataConstraint.WAIT );

	/* Class Methods... */
  public synchronized static int getArbitrarySubTaskIndex()
  {
    return ++ ARBITRARY_SUBTASK_INDEX;
  }



	/* Instance Variables */
  protected String taskname;
  protected Vector arguments;
  protected Vector constraints;
  protected int    extraIndent;


	/* Instance Methods */
  public DataSpawnTask()
  {
    taskname         = DataComponent.EMPTY_STRING;
    arguments        = new Vector();
    constraints      = new Vector();
    extraIndent      = 0;
  }

  public String getTaskName(                    )  { return taskname;        }
  public void   setTaskName( String theTaskname )  { taskname = theTaskname; }

  public Vector getArguments()   { return arguments;   }
  public Vector getConstraints() { return constraints; }

  public int    getConstraintCount() { return getConstraints() . count(); }


  public void   addArgument ( DataExpression theExpression )
  {
    getArguments() . addElement ( theExpression );
    theExpression . setParent ( this );
  }

  public void   addArgument ( DataExpression theExpression,
			      int            theAtIndex     )
  {
    getArguments() . insertElementAt ( theExpression, theAtIndex );
    theExpression . setParent ( this );
  }

  public void   addArgument ( String theArgument,
			      int    theAtIndex   )
  {
    DataExpression  dataExpression = new DataExpression();
    dataExpression . addSubcomponent ( theArgument );
    addArgument ( dataExpression, theAtIndex );
  }

  public void   removeArgument ( int theAtIndex )
  {
    getArguments() . removeElementAt ( theAtIndex );
  }


  public DataExpression getArgument ( int theIndex )
  {
    if ( ( theIndex >= 0 )   &&   ( theIndex < getArguments() . count() ) )
      return (DataExpression) ( getArguments() . elementAt ( theIndex ) ); 
    else
      return null;
  }

  public void   addConstraint ( DataConstraint theConstraint )
  {
    getConstraints() . addElement ( theConstraint );
    theConstraint . setParent ( this );
  }

  public DataConstraint getConstraint ( int theIndex )
  {
    if ( ( theIndex >= 0 )   &&   ( theIndex < getConstraints() . count() ) )
      return (DataConstraint) ( getConstraints() . elementAt ( theIndex ) ); 
    else
      return null;
  }

  public int  getExtraIndent ()               { return extraIndent; }
  public void setExtraIndent ( int theExtraIndent ) 
					      { extraIndent = theExtraIndent; }



	/* Overriden from DataStatement. */
	/* Includes the spawn-task-name when looking for names. */
  public boolean hasName ( String theName )
  {
    if (   ( getTaskName() != null              )
	&& ( getTaskName() . equals ( theName ) ) )
      return true;
    else
      return super.hasName ( theName );
  }


	/* Allow searching for contained expression subparts... */
  public boolean hasSubcomponentFraction ( String theString )
  {
    int i;

    if ( super . hasSubcomponentFraction ( theString ) )
      return true;

    for ( i=0;   i < getArguments() . count();   i++ )
      if ( getArgument ( i ) . hasSubcomponentFraction ( theString ) )
	return true;

    for ( i=0;   i < getConstraintCount();   i++ )
      if ( getConstraint ( i ) . hasSubcomponentFraction ( theString ) )
	return true;

    return false;
  }

	/* Allow searching of our expressions... */
  public void runOnSubcomponentFraction (
			String                      theString,
			RunOnSubcomponentInterface  theRunOnSubcomponentObject,
			Object                      theArgumentObject )
  {
    int i;

    super . runOnSubcomponentFraction ( theString,
					theRunOnSubcomponentObject,
					theArgumentObject );

    for ( i=0;   i < getArguments() . count();   i++ )
      getArgument ( i )
	. runOnSubcomponentFraction ( theString,
				      theRunOnSubcomponentObject,
				      theArgumentObject );


    for ( i=0;   i < getConstraintCount();   i++ )
      getConstraint ( i )
	. runOnSubcomponentFraction ( theString,
				      theRunOnSubcomponentObject,
				      theArgumentObject );
  }


  public DataComponent newComponent ( )
  {
    DataSpawnTask  spawnTask = new DataSpawnTask();

    spawnTask . addSubcomponent ( "\n" );
    spawnTask . setIndex ( DataSpawnTask.SPAWN );
    spawnTask . addSubcomponent ( " " );
    spawnTask . setIndex ( DataSpawnTask.TASK_NAME_INDEX );
    spawnTask . addSubcomponent ( " " );
    spawnTask . setIndex ( DataSpawnTask.OPEN_PAREN );
    spawnTask . setIndex ( DataSpawnTask.CLOSE_PAREN );
    spawnTask . addSubcomponent ( " " );
    spawnTask . setIndex ( DataSpawnTask.WITH );
    spawnTask . addSubcomponent ( " " );
    spawnTask . setIndex ( DataSpawnTask.SEMICOLON );

    return spawnTask;
  }


	/* Validates code that is outside of any Task */
  public void validateExternalCode( int                         theReference,
				    DataValidateCodeReturnValue theReturnValue)
    throws CompilationException
  {
	/* Validate our constraints */
    for ( int i=0;  i < getConstraints().count();  i++ )
    {
      getConstraint ( i )
	. validateExternalCode( theReference, theReturnValue );
    }

	/* Check for overriden constraints */
    DataStatement.validateAnyOverridenConstraints ( this, theReturnValue );
  }

	/* Validates code that is inside of a Task */
  public void validateInternalCode( int                         theReference,
				    DataValidateCodeReturnValue theReturnValue)
    throws CompilationException
  {
	/* Perform super-class validation */
    super . validateInternalCode ( theReference, theReturnValue );


	/* Validate our constraints */
    for ( int i=0;  i < getConstraints() . count();  i++ )
    {
      getConstraint ( i )
	. validateInternalCode ( theReference, theReturnValue );

      if ( getConstraint( i ) . getConstraintType() == DataConstraint.PARALLEL)
      {
	theReturnValue
	  . addWarning ( getConstraint ( i ) )
	  . write ( "PARALLEL constraint in a SPAWN statement WITH clause "
		    + "is meaningless.\n" );
      }
    }

        /* Validate mutually inconsistent constraints */
    DataStatement.validateInternalConstrainedObject ( theReference,
						      theReturnValue,
						      this );

	/* Check for overriden constraints */
    DataStatement.validateAnyOverridenConstraints ( this, theReturnValue );
  }



  public String getWarnString ( int      theObjectSubset,
				boolean  theIsGenerating )
  {
    return super . getWarnString ( theObjectSubset, theIsGenerating )
      + " or DataSpawnTask.TASK_ONLY (" + DataSpawnTask.TASK_ONLY
      + ") or DataSpawnTask.CONSTRAINTS_ONLY ("
      + DataSpawnTask.CONSTRAINTS_ONLY + ")";
  }

  public boolean isValidObjectSubset ( int      theObjectSubset,
				       boolean  theIsGenerating )
  {
    if (   ( theObjectSubset == DataSpawnTask.TASK_ONLY        )
	|| ( theObjectSubset == DataSpawnTask.CONSTRAINTS_ONLY ) )
      return true;
    else
      return super . isValidObjectSubset ( theObjectSubset, theIsGenerating );
  }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );

    if ( theObjectSubsetToParse == DataSpawnTask.TASK_ONLY )
      return theTDLParser . parseSpawnTask_TaskOnly ( null, 0 );

    else if ( theObjectSubsetToParse == DataSpawnTask.CONSTRAINTS_ONLY )
      return theTDLParser . parseSpawnTask_ConstraintsOnly ( null, 0 );

    else
      return theTDLParser . parseSpawnTask ( 0 );
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    if ( /*NOT*/ !  ( theObjectToCopy instanceof DataSpawnTask ) )
    {
      System.err.println ( "[DataSpawnTask:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataSpawnTask  theSpawnTaskToCopy = (DataSpawnTask) theObjectToCopy;
    DataComponent  aObject;
    int            i;


    if ( theObjectSubsetToSet != DataSpawnTask.CONSTRAINTS_ONLY )
    {
      taskname    = theSpawnTaskToCopy . taskname;

      if ( theCanDestroyObject )
      {
	arguments = theSpawnTaskToCopy . arguments;
      }
      else
      {
	arguments . removeAllElements();
	for ( i=0;  i < theSpawnTaskToCopy.arguments . count();  i++ )
	{
	  aObject
	    = (DataComponent) (theSpawnTaskToCopy.arguments . elementAt ( i ));
	  aObject = aObject . copyComponent();
	  if ( aObject == null )
	  {
	    System.err.println ( "[DataSpawnTask:setFrom]  ERROR:  "
			    + "Unable to replicate argument object:  "
			    + theSpawnTaskToCopy.arguments . elementAt ( i ) );
	    return false;
	  }
	  arguments . addElement ( aObject );
	} /* for ( 0 <= i < theSpawnTaskToCopy.arguments . count() ) */
      } /* if ( theCanDestroyObject ) ... else ... */
    } /* if ( theObjectSubsetToSet != DataSpawnTask.CONSTRAINTS_ONLY ) */


    if ( theObjectSubsetToSet != DataSpawnTask.TASK_ONLY )
    {
      if ( theCanDestroyObject )
      {
	constraints = theSpawnTaskToCopy . constraints;
      }
      else
      {
	constraints . removeAllElements();
	for ( i=0;  i < theSpawnTaskToCopy.constraints . count();  i++ )
	{
	  aObject
	    = (DataComponent) (theSpawnTaskToCopy.constraints . elementAt (i));
	  aObject = aObject . copyComponent();
	  if ( aObject == null )
	  {
	    System.err.println ( "[DataSpawnTask:setFrom]  ERROR:  "
			    + "Unable to replicate constraint object:  "
			    + theSpawnTaskToCopy.constraints . elementAt (i) );
	    return false;
	  }
	  constraints . addElement ( aObject );
	} /* for ( 0 <= i < theSpawnTaskToCopy.constraints . count() ) */
      } /* if ( theCanDestroyObject ) ... else ... */
    } /* if ( theObjectSubsetToSet != DataSpawnTask.TASK_ONLY ) */


    if ( theObjectSubsetToSet == DataSpawnTask.TASK_ONLY )
    {
      int    oldCloseParenIndex = getIndex ( DataSpawnTask.CLOSE_PAREN ),
	     spawnIndex         = getIndex ( DataSpawnTask.SPAWN       );

      for ( i=spawnIndex; i<oldCloseParenIndex; i++ )
	removeSubcomponent ( spawnIndex );

      for ( i = theSpawnTaskToCopy . getIndex ( DataSpawnTask.CLOSE_PAREN )
	        - 1;
	    i >= theSpawnTaskToCopy . getIndex ( DataSpawnTask.SPAWN );
	    i -- )
      {
	if ( theSpawnTaskToCopy . isSubcomponentAString ( i ) )
	  addSubcomponent ( theSpawnTaskToCopy
			      . getStringSubcomponent ( i ),
			    spawnIndex );
	else
	{
	  aObject = theSpawnTaskToCopy . getDataComponentSubcomponent ( i );

	  if ( theCanDestroyObject == false )
	  {
	    aObject = aObject . copyComponent();
	    if ( aObject == null )
	    {
	      System.err.println ( "[DataSpawnTask:setFrom]  ERROR:  "
		   + "Unable to replicate subcomponent object:  "
		   + theSpawnTaskToCopy . getDataComponentSubcomponent ( i ) );
	      return false;
	    }
	  }
	  addSubcomponent ( aObject, spawnIndex );
	}
      }

      setIndex ( DataSpawnTask.TASK_NAME_INDEX,
		 spawnIndex + 
		 theSpawnTaskToCopy.getIndex (DataSpawnTask.TASK_NAME_INDEX)
		 - theSpawnTaskToCopy.getIndex ( DataSpawnTask.SPAWN ) );

      setIndex ( DataSpawnTask.OPEN_PAREN,
		 spawnIndex + 
		 theSpawnTaskToCopy.getIndex ( DataSpawnTask.OPEN_PAREN )
		 - theSpawnTaskToCopy.getIndex ( DataSpawnTask.SPAWN ) );

      setIndex ( DataSpawnTask.CLOSE_PAREN,
		 spawnIndex + 
		 theSpawnTaskToCopy.getIndex ( DataSpawnTask.CLOSE_PAREN )
		 - theSpawnTaskToCopy.getIndex ( DataSpawnTask.SPAWN ) );
      return true;
    }
    
    else if ( theObjectSubsetToSet == DataSpawnTask.CONSTRAINTS_ONLY )
    {
      return true;
    }

    else
    {
      extraIndent = theSpawnTaskToCopy . extraIndent;

      return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			       theCanDestroyObject );
    }
  }


  public void generateTask ( DataDestination  theOutputDestination,
			     int              theObjectSubsetToGenerate,
			     boolean          theIsFollowingSpawn,
			     boolean          thePrintPreTaskNameSubcomponents,
			     boolean          theCorrectIndentation,
			     boolean          theStartIndentationOnNextLine,
			     boolean          theIsCxxSubsetGeneration )
  {
    int      i, indent = 0;
    boolean  needsToUndoLineMacros = false;


    if ( thePrintPreTaskNameSubcomponents == true )
    {
	/* Initialize us to generate non-significant tokens... */
      initializeGenerateSubcomponentIndex ( DataSpawnTask.SPAWN );

	/* Write any pre-task-name non-significant tokens */
      generateSubcomponents ( DataSpawnTask.TASK_NAME_INDEX,
			      theOutputDestination,
			      theObjectSubsetToGenerate, theIsFollowingSpawn );
    }
    else
    {
	/* Initialize us to generate non-significant tokens... */
      initializeGenerateSubcomponentIndex ( DataSpawnTask.TASK_NAME_INDEX );
    }

	/* Write our task name */
    theOutputDestination . write ( getTaskName() );

	/* Write any pre-"(" non-significant tokens */
    generateSubcomponents ( DataSpawnTask.OPEN_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write "(" */
    theOutputDestination . write ( DataSpawnTask.OPEN_PAREN );


	/* Are we trying to compensate for non-standard argument indentation?*/
    if ( theCorrectIndentation == true )
    {
      if ( theStartIndentationOnNextLine == true )
	theOutputDestination . addIndent ( DataComponent.STANDARD_INDENT );
      else
      {
	theOutputDestination . write ( " " ); /* Add a space in here */
	indent = theOutputDestination . indentToCurrentColumn();
      }
    }


	/* Write arguments */
    for ( i=0;  i < getArguments().count();  i++ )
    {
	/* Deal with #line macros */
      if (   ( theIsCxxSubsetGeneration                     == true )
	  && ( getArgument ( i ) . hasValidLineNumber()     == true )
	  && ( theOutputDestination . getEnableLineMacros() == true ) )
      {
	theOutputDestination . setUsingTdlFileName ( true );
	    /* Note: If ( theCorrectIndentation == true ),              *
	     *       adjust for skipping leading whitespace (newlines). */
	theOutputDestination
	  . makeNextLineNumber(  getArgument ( i ) . getLineNumber()
			       + (   ( theCorrectIndentation == true )
				   ? getArgument ( i )
				       . countLeadingSubcomponentNewlines()
				   : 0 ) );
	theOutputDestination . write ( "\n" ); /* Flush #line macro */

	needsToUndoLineMacros = true;
      }
      else if ( theCorrectIndentation == true )
      {
	if (   ( i > 0                                 )
	    || ( theStartIndentationOnNextLine == true ) )
	{
	  theOutputDestination . write ( "\n" );
	}
      }

      if ( theCorrectIndentation == true )
      {
	theOutputDestination . setStripLeadingWhitespace();
	theOutputDestination . setPersistentlyStripLeadingSpaces ( true );
      }

      getArgument ( i ) . generate ( theOutputDestination,
				     theObjectSubsetToGenerate );

      if ( theCorrectIndentation == true )
      {
	theOutputDestination . setPersistentlyStripLeadingSpaces ( false );
      }

	/* If there are more arguments */
      if ( (i+1) < getArguments().count() )
      {
	    /* Write "," */
	theOutputDestination . write ( DataSpawnTask.COMMA );
      }

    } /* FOR ( i=0;  i < getArguments().count();  i++ ) */


	/* Deal with #line macros */
    if ( needsToUndoLineMacros == true )
    {
      theOutputDestination . setUsingTdlFileName ( false );
      theOutputDestination . write ( "\n" ); /* Flush #line macro */
    }


	/* Write any pre-")" non-significant tokens */
    generateSubcomponents ( DataSpawnTask.CLOSE_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write ")" */
    theOutputDestination . write ( DataSpawnTask.CLOSE_PAREN );


	/* Are we trying to compensate for non-standard argument indentation?*/
    if ( theCorrectIndentation == true )
    {
      if ( theStartIndentationOnNextLine == true )
	theOutputDestination . removeIndent ( DataComponent.STANDARD_INDENT );
      else
	theOutputDestination . removeIndent ( indent );
    }
  }


  public void generateConstraints ( DataDestination  theOutputDestination,
				    int              theObjectSubsetToGenerate)
  {
    int i;

	/* Write constraints */
    for ( i=0;  i < getConstraints().count();  i++ )
    {
      getConstraint ( i ) . generate ( theOutputDestination,
				       theObjectSubsetToGenerate );

	  /* If there are more constraints */
      if ( (i+1) < getConstraints().count() )
      {
	    /* Write "," */
	theOutputDestination . write ( DataSpawnTask.COMMA );
      }
    }
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
	/* Do any extra indenting that we may require... */
	/* (Typically used for indenting spawns outside of Task definitions).*/
    theOutputDestination . addIndent ( getExtraIndent() );

       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

	/* Are we just doing the TASK_ONLY ? */
    if ( theObjectSubsetToGenerate == DataSpawnTask.TASK_ONLY )
    {
      generateTask ( theOutputDestination,
		     DataComponent.ENTIRE_OBJECT,
		     false,  /* No extra leading space   */
		     true,   /* Include leading comments */
		     false,  /* No extra indentation.    */
		     false,  /* No extra indentation.    */
		     isCxxSubset ( theObjectSubsetToGenerate ) );

		/* Stop any extra indenting... */
      theOutputDestination . removeIndent ( getExtraIndent() );
      return;
    }

	/* Are we just doing the CONSTRAINTS_ONLY ? */
    if ( theObjectSubsetToGenerate == DataSpawnTask.CONSTRAINTS_ONLY )
    {
      generateConstraints ( theOutputDestination, DataComponent.ENTIRE_OBJECT);
		/* Stop any extra indenting... */
      theOutputDestination . removeIndent ( getExtraIndent() );
      return;
    }

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex();

	/* Generate any labels that we have... */
    generateLabels ( theOutputDestination,
		     theObjectSubsetToGenerate,
		     (   isCxxSubset ( theObjectSubsetToGenerate )
		       ? DataLabelStatement.COMMENT_OUT_ID_LABELS
		       : DataComponent.ENTIRE_OBJECT ),
		     DataSpawnTask.SPAWN );

	/* Write any pre-spawn non-significant tokens */
    generateSubcomponents ( DataSpawnTask.SPAWN,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* If we are doing CXX generation... */
    if ( isCxxSubset ( theObjectSubsetToGenerate ) )
    {
      theOutputDestination . setNewlineText ( "    // " );
      theOutputDestination . write          ( "    // " );
    }

	/* Write our spawn keyword */
    theOutputDestination . write ( DataSpawnTask.SPAWN );


	/* Write our task-name & arguments... */
    generateTask ( theOutputDestination,
		   theObjectSubsetToGenerate,
		   true,    /* Add extra leading space if necessary. */
		   true,    /* Include leading comments.             */
		   false,   /* No extra indentation.                 */
		   false,   /* No extra indentation.                 */
		   false ); /* Disable special CXX generation code.  */

	/* Do we have constraints? */
    if ( getConstraints() . count() > 0 )
    {
	/* Write any pre-WITH non-significant tokens */
      generateSubcomponents ( DataSpawnTask.WITH,
			      theOutputDestination,
			      theObjectSubsetToGenerate, true );

	/* Write our WITH keyword */
      theOutputDestination . write ( DataSpawnTask.WITH );

	/* Write the constraints */
      generateConstraints ( theOutputDestination, DataComponent.ENTIRE_OBJECT);

    } /* IF ( we have constraints ) */

	/* Write any pre-";" non-significant tokens */
    generateSubcomponents ( DataSpawnTask.SEMICOLON,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write ";" */
    theOutputDestination . write ( DataSpawnTask.SEMICOLON );


	/* Write any remaining non-significant tokens */
	/* (There *SHOULD* *NOT* be any...  But just in case...) */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );

	/* If we are doing CXX generation... */
    if ( isCxxSubset ( theObjectSubsetToGenerate ) )
    {
      theOutputDestination . clearNewlineText();
      theOutputDestination . write ( "\n" );
      generateCxx ( theOutputDestination, theObjectSubsetToGenerate );
    }

		/* Stop any extra indenting... */
    theOutputDestination . removeIndent ( getExtraIndent() );
  }



  public void generateCxx ( DataDestination  theOutputDestination,
			    int              theSubsetToProduce   )
    throws CompilationException
  {
    int                  i, argIndent, invokeIndent;
    DataTaskDefinition   task;
    DataConstraint       waitConstraint  = null;
    String               spawnIdentifier;


	/* Deal with #line macros */
    theOutputDestination . setUsingTdlFileName ( false );
    theOutputDestination . write ( "\n" ); /* This newline will be skipped.*/


	/* Find our top-most object -- the Task that we are a part of */
    task = getParentTaskDefinition();

	/* If we don't have a parent */
    if ( task == null )
    {
	/* Error check */
      for ( i=0;  i < getConstraints() . count();  i++ )
      {
	   /* If we need a parent... Ie: have a constrainer-tag */
	if ( getConstraint ( i ) . getHasNonStandardEventTagTask() )
	{
	  throw new CompilationException (
	       "Internal Error in [DataSpawnTask:generateCxx]:  "
	       + "Spawn statement is not contained in a task.  "
	       + "Unable to satisfy constraint \""
	       + getConstraint ( i ) . toString()
	       + "\"." );
	}
      }

      DataSpawnTask.generateCxxOutsideOfTask ( theOutputDestination,
					       theSubsetToProduce,
					       this );
      return;
    }
  

	/* Find our default spawn-identifier */
    spawnIdentifier = task.getIdentifierForSubtask ( this );


	/** Deal with any switch-idents... **/
    theOutputDestination . enableSecondaryIndent();


	/***************************************/
	/* Begin (massive) invocation the Task */
	/***************************************/

	/* We need to establish any ON-AGENT constraints before the
	 * SPAWN statement, due to the TCM requirements that the AGENT
	 * be specified when the TCM-node is constructed, and the TDL aspect
	 * of delayed node creation.
	 */
    for ( i = getConstraints() . count() - 1;   i >= 0;   i-- )
    {
      if ( getConstraint (i) . getConstraintType() == DataConstraint.ON_AGENT )
	break;
    }


	/* Note:  do-start-invoking must be a complete stand-alone statement,
	 * Otherwise we can't access the current Task from the handle-manager
	 */
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_INSTANCE );
    theOutputDestination . write ( " " );

    theOutputDestination . write ( ". " );
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_DO_START_INVOKING );
    theOutputDestination . write ( " ( " );
    argIndent = theOutputDestination . indentToCurrentColumn();
    theOutputDestination . write ( "\"" );
    theOutputDestination . write ( spawnIdentifier );
    theOutputDestination . write ( "\"" );
	/* Add the ON-AGENT constraint */
    if ( i >= 0 )
    {
      theOutputDestination . write ( ",\n" );
      getConstraint ( i ) . generateCxxTaskInternal ( theOutputDestination );
    }
    theOutputDestination . removeIndent ( argIndent );
    theOutputDestination . write ( " );\n" );



	/*********************/
	/* Handle Set-Action */
	/*********************/

	/* Write "task-spawn-manager . doSetAction ( " */
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_INSTANCE );
    theOutputDestination . write ( " " );
    invokeIndent = theOutputDestination . indentToCurrentColumn();

    theOutputDestination . write ( ". " );
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_DO_SET_ACTION );
    theOutputDestination . write ( " ( " );

    theOutputDestination
      . write ( DataComponent.CXX_CREATEACTION_FUNCTION_LEAD );
    theOutputDestination . write ( getTaskName() );
    theOutputDestination . write ( " ( " );

	/* Indent arguments... */
    argIndent = theOutputDestination . indentToCurrentColumn();

	/* Write:  "* ( _TDL_SpawnedTasks . getCurrentTask() )," */
    theOutputDestination . write ( "* ( " );    
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_INSTANCE );
    theOutputDestination . write ( " . " );
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_GET_CURRENT_TASK );
    theOutputDestination . write ( "() )" );

	  /* Write the task-argument-names as arguments to the setaction... */
    for ( i=0;  i < getArguments() . count();  i++ )
    {
	/* Deal with #line macros */
      theOutputDestination . setUsingTdlFileName ( true );
      theOutputDestination
	. makeNextLineNumber (   getArgument ( i ) . getLineNumber()
			       + getArgument ( i )
				   . countLeadingSubcomponentNewlines() );

	/* Write the argument */
      theOutputDestination . write ( ",\n" );
      theOutputDestination . setStripLeadingWhitespace();
      theOutputDestination . setPersistentlyStripLeadingSpaces ( true  );
      getArgument ( i ) . generate( theOutputDestination, theSubsetToProduce );
      theOutputDestination . setPersistentlyStripLeadingSpaces ( false );
    }

	/* Deal with #line macros */
    theOutputDestination . setUsingTdlFileName ( false );

    theOutputDestination . write ( " ) )\n" );


	/* Stop indenting Arguments... */
    theOutputDestination . removeIndent ( argIndent );
    


	/**************************************/
	/*** SPAWN Constraints, pre-insert  ***/
	/**************************************/

    for ( i=0;  i < getConstraints() . count();  i++ )
    {
      if (    DataConstraint.getIsPostInsertConstraint ( getConstraint ( i ) )
	   == false )
      {
		/* Skip overriden constraints */
	if (  ( DataConstraint.getIsConstraintOverriden ( getConstraint (i),
							  getConstraints()  ) )
		/* Skip ON-AGENT constraints */
	    || (    getConstraint ( i ) . getConstraintType()
		 == DataConstraint.ON_AGENT                                   )
	    )
	{
	  continue;
	}

	theOutputDestination . write ( ". " );
	theOutputDestination
	  . write ( DataComponent.CXX_TDL_TASK_HANDLER_DO_APPLY_CONSTRAINT );
	theOutputDestination . write ( " ( " );
	getConstraint ( i ) . generateCxxTaskInternal ( theOutputDestination );
	theOutputDestination . write ( " )\n" );
      }
    }



	/**********************/
	/* Handle insert-node */
	/**********************/

    theOutputDestination . write ( ". " );
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_DO_INSERT );
    theOutputDestination . write ( " ()\n" );



	/***************************************/
	/*** SPAWN Constraints, post-insert  ***/
	/***************************************/

    for ( i=0;  i < getConstraints() . count();  i++ )
    {
      if (    DataConstraint.getIsPostInsertConstraint ( getConstraint ( i ) )
	   == true )
      {
		/* Skip overriden constraints */
	if ( DataConstraint.getIsConstraintOverriden ( getConstraint ( i ),
						       getConstraints()    ) )
	{
	  continue;
	}


	    /* No need to check for ON-AGENT constraints.  *
	     * They are *NEVER* post-insert constraints... */


	    /* Only do WAIT constraints once. */
	if ( getConstraint ( i ) . getConstraintType() == DataConstraint.WAIT )
	{
	  if ( waitConstraint != null )
	    continue;  /* Skip this constraint. */
	  else
	    waitConstraint = getConstraint ( i );
	}

	theOutputDestination . write ( ". " );
	theOutputDestination
	  . write ( DataComponent.CXX_TDL_TASK_HANDLER_DO_APPLY_CONSTRAINT );
	theOutputDestination . write ( " ( " );
	getConstraint ( i ) . generateCxxTaskInternal ( theOutputDestination );
	theOutputDestination . write ( " )\n" );
      }
    }


	/******************************/
	/* END the spawn statement... */
	/******************************/

    theOutputDestination . write ( ". " );
    theOutputDestination
      . write ( DataComponent.CXX_TDL_TASK_HANDLER_FINISH_INVOKING );
	/* Stop indenting now... */
    theOutputDestination . removeIndent ( invokeIndent );
    theOutputDestination . write ( "();\n" );


	/** Stop Dealing with any switch-idents... **/
    theOutputDestination . disableSecondaryIndent();


  } /* public void generateCxx ( ... ) */


  public static void generateCxxOutsideOfTask (
					  DataDestination theOutputDestination,
					  int             theSubsetToProduce,
					  DataSpawnTask   theDataSpawnTask    )
  {
    privateGenerateCxxOutsideOfTask ( theOutputDestination,
				      theSubsetToProduce,
				      theDataSpawnTask );
  }

  public static void generateCxxOutsideOfTask (
				     DataDestination    theOutputDestination,
				     int                theSubsetToProduce,
				     DataTaskDefinition theDataTaskDefinition )
  {
    privateGenerateCxxOutsideOfTask ( theOutputDestination,
				      theSubsetToProduce,
				      theDataTaskDefinition );
  }

	/** This method is static so that it can serve two purposes.
	  * (1) Generating the CXX version of SPAWNs outside of Tasks
	  * And (2), Generating the _TDL_SpawnAndWait_* function for
	  * DataTaskDefinition objects.
	  */
  protected static void privateGenerateCxxOutsideOfTask (
				         DataDestination  theOutputDestination,
					 int              theSubsetToProduce,
					 DataComponent    theTaskSource       )
  {
    DataConstraint     waitConstraint        = null;
    DataSpawnTask      theDataSpawnTask      = null;
    DataTaskDefinition theDataTaskDefinition = null;
    boolean            isDataSpawnTask;
    String             taskName, spawnObject, errorLocation, returnValueString;
    int                i, ifIndent, argIndent;

	/* Find what flavor we are generating... */
    if ( theTaskSource instanceof DataSpawnTask )
    {
      isDataSpawnTask  = true;
      theDataSpawnTask = (DataSpawnTask) theTaskSource;
    }

    else if ( theTaskSource instanceof DataTaskDefinition )
    {
      isDataSpawnTask  = false;
      theDataTaskDefinition = (DataTaskDefinition) theTaskSource;
    }

    else
    {
      System.err.println ( "[DataSpawnTask:privateGenerateCxxOutsideOfTask]  "
			   + "Programmer Error:  theTaskSource is not a "
			   + "DataSpawnTask or a DataTaskDefinition. "
			   + "Aborting..." );
      return;
    }
    

	/* Choose our spawn-object-name... */
    if ( isDataSpawnTask )
    {
      taskName      =  theDataSpawnTask . getTaskName();
      spawnObject   = (  DataComponent.CXX_TDL_ARBITRARY_SPAWN_TASK_NAME_LEAD
		       + DataSpawnTask.getArbitrarySubTaskIndex() );
      errorLocation = "\"[:\'" + taskName + "\' - " + spawnObject + "]\"";
      returnValueString = "";
    }
    else
    {
      taskName      = theDataTaskDefinition . getTaskName();
      spawnObject   = DataComponent.CXX_SPAWNWAIT_CHILD_NODE_NAME;
      errorLocation = ( "\"[" + DataComponent.CXX_SPAWNWAIT_FUNCTION_LEAD
			+ taskName + "]\"" );
      returnValueString = DataComponent.CXX_SPAWNWAIT_RETURN_VARIABLE + " = ";

	/* Write "return-value-type  returnValue;" */
      theOutputDestination
        . write ( DataComponent.CXX_SPAWNWAIT_RETURN_VALUE );
      theOutputDestination . write ( "    " );
      theOutputDestination
	. write ( DataComponent.CXX_SPAWNWAIT_RETURN_VARIABLE );
      theOutputDestination . write ( " = " );
      theOutputDestination . write ( DataComponent.CXX_TCM_ERROR );
      theOutputDestination . write ( ";\n" );
    }

	/* Declare/Allocate the spawnObject */
    theOutputDestination . write ( DataComponent.CXX_ALLOCATE_RETURN_VALUE );
    theOutputDestination . write ( "  " );
    theOutputDestination . write ( spawnObject );
    theOutputDestination . write ( " = " );
    theOutputDestination . write ( DataComponent.CXX_ALLOCATE_FUNCTION_LEAD);
    theOutputDestination . write ( taskName );
    theOutputDestination . write ( " ( \"" );
    theOutputDestination . write ( taskName );

	/* Identify whether this is a outside-Task or a auto-Task */
    if ( isDataSpawnTask )
      theOutputDestination . write ( "-outsideTask\" );\n" );
    else
      theOutputDestination . write ( "-auto,wait\" );\n" );
  


	/*******************************************/
	/* Some Constraints, such as ON-TERMINATE, */
	/* need to allocate secondary subtasks.    */
	/*******************************************/

    if ( isDataSpawnTask )
    {
      for ( i=0;  i < theDataSpawnTask . getConstraintCount();  i++ )
      {
	theDataSpawnTask
	  . getConstraint ( i )
	  . generateCxxExternalSubtaskDeclarations ( theOutputDestination );
      }
    }

    theOutputDestination . write ( "\n" );




	/**********************************/
	/* Begin the massive IF statement */
	/**********************************/

    theOutputDestination . write ( "if (" );
    
	  /* Indent Master if-statement expression */
    ifIndent = theOutputDestination . indentToCurrentColumn();


	/***************************/
	/* Handle allocation check */
	/***************************/
    theOutputDestination . write ( "   " );
    theOutputDestination . write ( DataComponent.CXX_VERIFY_ALLOCATION );
    theOutputDestination . write ( " ( " );
    
	  /* Indent arguments... */
    argIndent = theOutputDestination . indentToCurrentColumn();

    theOutputDestination . write ( errorLocation );
    theOutputDestination . write ( ",\n" );

    theOutputDestination . write ( spawnObject );
    theOutputDestination . write ( " )\n" );

	/* Stop indenting Arguments... */
    theOutputDestination . removeIndent ( argIndent );



	/*********************/
	/* Handle set-action */
	/*********************/

    theOutputDestination . write ( "\n&& " );
    theOutputDestination . write ( DataComponent.CXX_VERIFY_SET_ACTION );
    theOutputDestination . write ( " ( " );
    
	  /* Indent arguments... */
    argIndent = theOutputDestination . indentToCurrentColumn();

    theOutputDestination . write ( errorLocation );
    theOutputDestination . write ( ",\n" );

    theOutputDestination . write ( returnValueString );

    theOutputDestination . write ( DataComponent.CXX_TCM_SET_ACTION_FUNCTION );
    theOutputDestination . write ( " ( " );

	/* Indent next arguments to this column ... */
    theOutputDestination . removeIndent ( argIndent );
    argIndent = theOutputDestination . indentToCurrentColumn();

    theOutputDestination . write ( spawnObject );
    theOutputDestination . write ( " ,\n" );

    theOutputDestination
      . write ( DataComponent.CXX_CREATEACTION_FUNCTION_LEAD );
    theOutputDestination . write ( taskName );
    theOutputDestination . write ( " ( " );


	/* Indent remaining arguments to this column ... */
    theOutputDestination . removeIndent ( argIndent );
    argIndent = theOutputDestination . indentToCurrentColumn();

	/* Write the spawnObject (TCM_Task_Tree_Ref) */
    theOutputDestination . write ( spawnObject );      

	  /* Write the task-argument-names as arguments to the setaction... */
    if ( isDataSpawnTask )
    {
      for ( i=0;  i < theDataSpawnTask . getArguments() . count();  i++ )
      {
	    /* Deal with #line macros */
	theOutputDestination . setUsingTdlFileName ( true );
	theOutputDestination . makeNextLineNumber (
	       theDataSpawnTask . getArgument ( i ) . getLineNumber()
	     + theDataSpawnTask . getArgument ( i )
	         . countLeadingSubcomponentNewlines() );

	    /* Write the argument */
	theOutputDestination . write ( ",\n" );
	theOutputDestination . setStripLeadingWhitespace();
	theOutputDestination . setPersistentlyStripLeadingSpaces ( true  );
	theDataSpawnTask . getArgument ( i ) . generate ( theOutputDestination,
							  theSubsetToProduce );
	theOutputDestination . setPersistentlyStripLeadingSpaces ( false );
      }

      	/* Deal with #line macros */
      theOutputDestination . setUsingTdlFileName ( false );
    }
    else
    {
      for ( i=0;  i < theDataTaskDefinition . getTaskArgumentCount();  i++ )
      {
	theOutputDestination . write ( ",\n" );
	theOutputDestination . setStripLeadingWhitespace();
	theOutputDestination
	  . write ( theDataTaskDefinition . getTaskArgument ( i )
				          . getArgumentName()     );
      }
    }

    theOutputDestination . write ( " ) ) )\n" );

	/* Stop indenting Arguments... */
    theOutputDestination . removeIndent ( argIndent );



	/**************************************/
	/*** SPAWN Constraints, pre-insert  ***/
	/**************************************/
    if ( isDataSpawnTask )
    {
      for ( i=0;  i < theDataSpawnTask . getConstraints() . count();  i++ )
      {
	if (    DataConstraint.getIsPostInsertConstraint (
				      theDataSpawnTask . getConstraint ( i )  )
	     == false )
	{
		/* Skip overriden constraints */
	  if ( DataConstraint.getIsConstraintOverriden (
				     theDataSpawnTask . getConstraint ( i ),
				     theDataSpawnTask . getConstraints()    ) )
	  {
	    continue;
	  }

	  theDataSpawnTask . getConstraint ( i )
			   . generateCxxTaskExternal ( theOutputDestination,
						       errorLocation,
						       spawnObject,
						       returnValueString,
						       true );
	}
      }
    }



	/**********************/
	/* Handle insert-node */
	/**********************/

    theOutputDestination . write ( "\n&& " );
    theOutputDestination . write ( DataComponent.CXX_VERIFY_INSERT_NODE );
    theOutputDestination . write ( " ( " );
    
	  /* Indent arguments... */
    argIndent = theOutputDestination . indentToCurrentColumn();

    theOutputDestination . write ( errorLocation );
    theOutputDestination . write ( ",\n" );

    theOutputDestination . write ( returnValueString );
    theOutputDestination . write ( DataComponent.CXX_TCM_INSERT_NODE_FUNCTION);
    theOutputDestination . write ( " ( " );

	/* Indent remaining arguments to this column ... */
    theOutputDestination . removeIndent ( argIndent );
    argIndent = theOutputDestination . indentToCurrentColumn();

	/* When spawning Outside-of-Tasks, we use the parent of TCM-ROOT-NODE.
	 * But for generating _TDL_SpawnAndWait_* functions, this is specified.
	 */
    if ( isDataSpawnTask )
      theOutputDestination . write ( DataComponent.CXX_TCM_ROOT_NODE );
    else
      theOutputDestination . write (DataComponent.CXX_SPAWNWAIT_ARGUMENT_NAME);

    theOutputDestination . write ( ",\n" );
    theOutputDestination . write ( spawnObject );
    theOutputDestination . write ( ",\n" );
    if ( isDataSpawnTask )
      theOutputDestination . write ( "TRUE" );
    else
      theOutputDestination . write ( "FALSE" );
    theOutputDestination . write ( " ) )\n" );

	/* Stop indenting Arguments... */
    theOutputDestination . removeIndent ( argIndent );



	/***************************************/
	/*** SPAWN Constraints, post-insert  ***/
	/***************************************/
    if ( isDataSpawnTask )
    {
      for ( i=0;  i < theDataSpawnTask . getConstraints() . count();  i++ )
      {
	if (    DataConstraint.getIsPostInsertConstraint (
				       theDataSpawnTask . getConstraint ( i ) )
	     == true )
	{
		/* Skip overriden constraints */
	  if ( DataConstraint.getIsConstraintOverriden (
				     theDataSpawnTask . getConstraint ( i ),
				     theDataSpawnTask . getConstraints()    ) )
	  {
	    continue;
	  }

		/* Only do WAIT constraints once. */
	  if (    theDataSpawnTask . getConstraint ( i ) . getConstraintType()
	       == DataConstraint.WAIT )
	  {
	    if ( waitConstraint != null )
	      continue;  /* Skip this constraint. */
	    else
	      waitConstraint = theDataSpawnTask . getConstraint ( i );
	  }

	  theDataSpawnTask . getConstraint ( i )
			   . generateCxxTaskExternal ( theOutputDestination,
						       errorLocation,
						       spawnObject,
						       returnValueString,
						       true );
	}
      }
    }


	/******************************************/
	/* Generate WAIT constraint, if necessary */
	/******************************************/

    if (   ( waitConstraint  == null  )
	&& ( isDataSpawnTask == false ) )
    {
      DataSpawnTask.WAIT_CONSTRAINT
	. generateCxxTaskExternal ( theOutputDestination,
				    errorLocation,
				    spawnObject,
				    returnValueString,
				    true );
    }



	/**************************************/
	/* Terminate the massive IF statement */
	/**************************************/

    theOutputDestination . write ( ")" );

	/* Stop indenting the Master if-statement expression */
    theOutputDestination . removeIndent ( ifIndent  );

	/* Write the body. */
    theOutputDestination . write        ( "\n{\n" );
    theOutputDestination . addIndent    ( DataComponent.STANDARD_INDENT );
    theOutputDestination . write        ( 
	"; /* Null Statement!  (Task has been successfully SPAWN'ed) */\n" );
    theOutputDestination . removeIndent ( DataComponent.STANDARD_INDENT );
    theOutputDestination . write        ( "}\n" );



	/***************************************/
	/* Write the final return statement... */
        /***************************************/
    if ( isDataSpawnTask == false )
    {
      theOutputDestination . write ( "\nreturn " );
      theOutputDestination
	. write ( DataComponent.CXX_SPAWNWAIT_RETURN_VARIABLE );
      theOutputDestination . write ( ";\n" );
    }

  } /* public void privateGenerateCxxOutsideOfTask ( ... ) */




  public boolean isSpawnRelatedStatement ( )
  {
    return true;
  }

  public int  getChildStatementCount()
  {
    return super . getChildStatementCount()
      + getConstraintChildStatementCount();
  }

  public DataStatement  getChildStatement ( int theIndex )
  {
    if ( theIndex < getConstraintChildStatementCount() )
      return getConstraintChildStatement ( theIndex );
    else
      return super . getChildStatement(   theIndex
					- getConstraintChildStatementCount() );
  }



	/********************************************************/
        /********** Methods to interface with Graphing **********/
	/********************************************************/
  public boolean supportsGraphType ( int theGraphType )
  {
    if (   ( theGraphType == GraphRoot.GUI_TREE_GRAPH ) )
      return true;
    else
      return false;
  }

  public GraphModule createGraphModule ( int theGraphType )
  {
    if (   ( theGraphType != GraphRoot.GUI_TREE_GRAPH ) )
      return null;

    /* else ... */

    /* Building:  +- spawn - [.]
     *            |    |
     *            |   with - [.]
     *            |
     */

    ScrollTextOfDataComponent   constraintText
      = new ScrollTextOfDataComponent ( this, DataSpawnTask.CONSTRAINTS_ONLY );

    GraphNode       constraintNode = new GraphNode ( constraintText,
					     GraphComponent.LEFT,
					     DataSpawnTask.SPAWN_COLOR,
					     true, null );
						     
    GraphConnection toConstraints = new GraphConnection ( GraphComponent.LEFT,
					     DataSpawnTask.SPAWN_COLOR,
					     false, false, constraintNode );

    GraphNode       withNode = new GraphNode ( new SimpleLabel ( "WITH" ),
					     GraphComponent.UP,
					     DataSpawnTask.SPAWN_COLOR,
					     false, toConstraints );

    GraphConnection toWith = new GraphConnection ( GraphComponent.UP,
					     DataSpawnTask.SPAWN_COLOR,
					     false, true, withNode );

    ScrollTextOfDataComponent   taskText
      = new ScrollTextOfDataComponent ( this, DataSpawnTask.TASK_ONLY );

    GraphNode       taskNode = new GraphNode ( taskText,
					     GraphComponent.LEFT,
					     DataSpawnTask.SPAWN_COLOR,
					     true, null );

    GraphConnection toTask = new GraphConnection ( GraphComponent.LEFT,
					     DataSpawnTask.SPAWN_COLOR,
					     false, false, taskNode );

    GraphNode       spawnNode = new GraphNode ( new SimpleLabel ( "SPAWN" ),
					     GraphComponent.LEFT,
					     DataSpawnTask.SPAWN_COLOR,
					     false, toTask );

    spawnNode . addChild ( toWith );


    GraphConnection toSpawn = new GraphConnection ( GraphComponent.LEFT,
					     DataSpawnTask.SPAWN_COLOR,
					     false, false, spawnNode );

    GraphMeet       startMeet = new GraphMeet ( GraphComponent.UP,
					     DataSpawnTask.SPAWN_COLOR,
					     GraphComponent.LEFT,
					     toSpawn );

    GraphConnection tail = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );
    startMeet . addChild ( tail );

    return new GraphModule ( startMeet, tail, null, null,
			     this, true, theGraphType );
  }

	/* Invoked by this.update(*) invoking GraphModule.update() */
  public boolean updateGraphModule ( DataComponent  theModifiedComponent,
				     Object         theStoredData,
				     GraphModule    theGraphModule )
  {
	/* There is nothing to do here.. */
    return true;
  }

}

