
import netscape.application.Color;

/*
 * This represents a C "while" statement.
 */

public class DataWhileStatement extends DataIterationStatement
{
	/* Class variables */
  public final static String   WHILE            = "while";
  public final static String   OPEN_PAREN       = "(";
  public final static String   CLOSE_PAREN      = ")";
  public final static String   EXPRESSION_INDEX = "ExpressionIndex";
  public final static String   STATEMENT_INDEX  = "StatementIndex";
  public final static String   SEMICOLON        = ";";

  public       static Color    WHILE_COLOR           = Color.magenta;
  public final static int      WHILE_EXPRESSION_ONLY = 6;


	/* Instance Variables */
  protected DataExpression          expression;


	/* Instance Methods */
  public DataWhileStatement()
  {
    expression = null;
  }

  public DataExpression getExpression () { return expression; }

  public void           setExpression ( DataExpression theExpression )
  {
    if ( expression != null )
      expression . setParent ( null );

    expression = theExpression;

    if ( expression != null )
      expression . setParent ( this );
  }


	/* Allow searching of our expression... */
  public boolean hasSubcomponentFraction ( String theString )
  {
    return super               . hasSubcomponentFraction ( theString )
      || (   ( getExpression() != null )
	  && ( getExpression() . hasSubcomponentFraction ( theString ) ) );
  }

	/* Allow searching of our expressions... */
  public void runOnSubcomponentFraction (
			String                      theString,
			RunOnSubcomponentInterface  theRunOnSubcomponentObject,
			Object                      theArgumentObject )
  {
    super . runOnSubcomponentFraction ( theString,
					theRunOnSubcomponentObject,
					theArgumentObject );

    if ( getExpression() != null )
      getExpression() . runOnSubcomponentFraction ( theString,
						    theRunOnSubcomponentObject,
						    theArgumentObject );
  }


  public DataComponent newComponent ( )
  {
    DataWhileStatement  whileStatement = new DataWhileStatement();

    whileStatement . addSubcomponent ( "\n" );
    whileStatement . setIndex ( DataWhileStatement.WHILE );
    whileStatement . addSubcomponent ( " " );
    whileStatement . setIndex ( DataWhileStatement.OPEN_PAREN  );
    whileStatement . setIndex ( DataWhileStatement.CLOSE_PAREN );

    return whileStatement;
  }


  public String getWarnString ( int      theObjectSubset,
				boolean  theIsGenerating )
  {
    return super . getWarnString ( theObjectSubset, theIsGenerating )
      + " or DataWhileStatement.WHILE_EXPRESSION_ONLY ("
      + DataWhileStatement.WHILE_EXPRESSION_ONLY + ")";
  }

  public boolean isValidObjectSubset ( int      theObjectSubset,
				       boolean  theIsGenerating )
  {
    if ( theObjectSubset == DataWhileStatement.WHILE_EXPRESSION_ONLY )
      return true;
    else
      return super . isValidObjectSubset ( theObjectSubset, theIsGenerating );
  }


  public boolean isValid ( int theObjectSubsetToValidate )
  {
      /* We should check to make sure it parses later on. *DBA*/

    if ( theObjectSubsetToValidate == DataWhileStatement.WHILE_EXPRESSION_ONLY)
    {
      return getExpression() != null;
    }

	/* Simple checks that we can do here.. */
    if (   ( getExpression() == null )
	|| ( getStatement()  == null ) )
      return false;

    return super.isValid ( theObjectSubsetToValidate );
  }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );

    if ( theObjectSubsetToParse == DataWhileStatement.WHILE_EXPRESSION_ONLY )
      return theTDLParser . parseExpressionObject ( 0 );
    else
      return theTDLParser . parseWhile ( 0 );
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    DataComponent  aObject = theObjectToCopy;


    if (   ( theObjectSubsetToSet == DataWhileStatement.WHILE_EXPRESSION_ONLY )
	&& ( theObjectToCopy instanceof DataExpression ) )
    {
      if (   ( theCanDestroyObject == false )
	  && ( aObject             != null  ) )
      {
	aObject = aObject . copyComponent();
     
	if ( aObject == null )
	{
	  System.err.println ( "[DataWhileStatement:setFrom(EXPR)]  Error:  "
		         + "Unable to replicate object:  " + theObjectToCopy );
	  return false;
	}
      }
      setExpression ( (DataExpression) aObject );
      return true;
    }

    if ( /*NOT*/ !  ( theObjectToCopy instanceof DataWhileStatement ) )
    {
      System.err.println ( "[DataWhileStatement:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataWhileStatement  theWhileStatementToCopy
			      = (DataWhileStatement) theObjectToCopy;

    aObject = theWhileStatementToCopy . expression;

    if (   ( theCanDestroyObject == false )
	&& ( aObject             != null  ) )
    {
      aObject = aObject . copyComponent();

      if ( aObject == null )
      {
	System.err.println ( "[DataWhileStatement:setFrom]  Error:  "
		         + "Unable to replicate object:  " + theObjectToCopy );
	return false;
      }
    }

    setExpression ( (DataExpression) aObject );


    if ( theObjectSubsetToSet == DataWhileStatement.WHILE_EXPRESSION_ONLY )
      return true;

    return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject );
  }


  public void generateExpression ( DataDestination  theOutputDestination,
				   int              theObjectSubsetToGenerate,
				   boolean          theIsPartOfWhileStatement )
  {
    boolean  needsLeadingSpace  = theIsPartOfWhileStatement,
             needsTrailingSpace = theIsPartOfWhileStatement;

    if ( getExpression() != null )
    {
      if ( getExpression() . hasLeadingWhitespaceSubcomponent() )
	needsLeadingSpace = false;
      if ( getExpression() . hasTrailingWhitespaceSubcomponent() )
	needsTrailingSpace = false;
    }

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex ( DataWhileStatement.OPEN_PAREN );

	/* Write any pre-EXPRESSION non-significant tokens */
    generateSubcomponents ( DataWhileStatement.EXPRESSION_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate,
			    needsLeadingSpace );

	/* Write out expression, if we have one... */
    if ( getExpression() != null )
      getExpression() . generate ( theOutputDestination,
				   theObjectSubsetToGenerate );


	/* Write any pre-")" tokens */
    generateSubcomponents ( DataWhileStatement.CLOSE_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate,
			    needsTrailingSpace );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
    int i;

       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

		/* Handle subset case... */
    if ( theObjectSubsetToGenerate == DataWhileStatement.WHILE_EXPRESSION_ONLY)
    {
      generateExpression ( theOutputDestination, DataComponent.ENTIRE_OBJECT,
			   false );
      return;
    }

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex();

	/* Generate any labels that we have... */
    generateLabels ( theOutputDestination,
		     theObjectSubsetToGenerate,
		     DataWhileStatement.WHILE,
		     getLabel() == null /* Disable #line macros if no labels.*/
		    );

	/* Write any pre-while non-significant tokens */
    generateSubcomponents ( DataWhileStatement.WHILE,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Generate extra pre-iteration code. */
    generatePreIterationCode ( theOutputDestination,
			       theObjectSubsetToGenerate );

	/* Generate relevant #line-number macros */
    generateRelevantLineNumberMacros ( theOutputDestination,
				       getLineNumber(),
				       0 );

	/* Write our WHILE keyword */
    theOutputDestination . write ( DataWhileStatement.WHILE );

	/* Write any pre-"(" non-significant tokens */
    generateSubcomponents ( DataWhileStatement.OPEN_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write "(" */
    theOutputDestination . write ( DataWhileStatement.OPEN_PAREN );


	/* Write pre-Expression tokens, the expression, and pre-")" tokens */
    generateExpression ( theOutputDestination, theObjectSubsetToGenerate,
			 true );


	/* Write ")" */
    theOutputDestination . write ( DataWhileStatement.CLOSE_PAREN );


	/* Write any pre-STATEMENT non-significant tokens */
    generateSubcomponents ( DataWhileStatement.STATEMENT_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );


	/* Write the statement, if we have one... */
    if ( getStatement() != null )
    {
	/* Generate extra inside-iteration code */
      addInsideIterationCode ( theObjectSubsetToGenerate, getStatement() );

      getStatement() . generate ( theOutputDestination,
				  theObjectSubsetToGenerate );

	/* Remove extra inside-iteration code */
      removeInsideIterationCode ( theObjectSubsetToGenerate, getStatement() );
    }
    else  /* Try to recover gracefully if we have a problem... */
    {
	/* Generate empty body statement with extra inside-iteration code */
      generateEmptyIterationBodyStatement ( theOutputDestination,
					    theObjectSubsetToGenerate );
    }

	/* Generate extra post-iteration code. */
    generatePostIterationCode ( theOutputDestination,
				theObjectSubsetToGenerate );


	/* Write any remaining non-significant tokens */
	/* (There *SHOULD* *NOT* be any...  But just in case...) */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }



	/********************************************************/
        /********** Methods to interface with Graphing **********/
	/********************************************************/
  public boolean supportsGraphType ( int theGraphType )
  {
    if (   ( theGraphType == GraphRoot.GUI_TREE_GRAPH ) )
      return true;
    else
      return false;
  }

  public GraphModule createStatementGraphModule ( int theGraphType )
  {
    if (   ( theGraphType != GraphRoot.GUI_TREE_GRAPH ) )
      return null;

    /* else ... */

    /* Building:  +- while - "[.]"
     *            |    |
     */

    ScrollTextOfDataComponent  expressionText
      = new ScrollTextOfDataComponent ( this, DataWhileStatement
					       . WHILE_EXPRESSION_ONLY );

    GraphNode       expressionNode = new GraphNode ( expressionText,
					     GraphComponent.LEFT,
					     DataWhileStatement.WHILE_COLOR,
					     true, null );
						     
    GraphConnection toExpression = new GraphConnection ( GraphComponent.LEFT,
					     DataWhileStatement.WHILE_COLOR,
					     false, false, expressionNode );

    GraphNode       whileNode = new GraphNode ( new SimpleLabel( "WHILE" ),
					     GraphComponent.LEFT,
					     DataWhileStatement.WHILE_COLOR,
					     false, toExpression );

    GraphConnection toChildren = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );
    whileNode . addChild ( toChildren );


    GraphConnection toWhile = new GraphConnection ( GraphComponent.LEFT,
					     DataWhileStatement.WHILE_COLOR,
					     false, false, whileNode );

    GraphMeet       startMeet = new GraphMeet ( GraphComponent.UP,
					     DataWhileStatement.WHILE_COLOR,
					     GraphComponent.LEFT, toWhile );

    GraphConnection tail = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );
    startMeet . addChild ( tail );

    GraphModule  graphModule = new GraphModule ( startMeet, tail, toChildren,
						 null, this, true, 
						 theGraphType );

    updateGraphModule ( this, null, graphModule );

    return graphModule;
  }



	/* Invoked by this.update(*) invoking GraphModule.update() */
  public boolean updateGraphModule ( DataComponent  theModifiedComponent,
				     Object         theStoredData,
				     GraphModule    theGraphModule )
  {
	/* Skip updates to parents... */
    if ( isPartOfThisStatement ( theModifiedComponent ) == false )
      return true;

    if ( getStatement() != null )
      return getStatement() . updateGraphModule( theGraphModule,
						 GraphModule.CHILD_COMPONENT );
    else
      return true;
  }

}
