
import java.io.PrintStream;
import netscape.application.*;
import netscape.util.*;

/* scrolling, scaling, directed acyclic Graph  */
public class GraphView extends View implements ExtendedTarget, Target
{
	/* Class Constants */
  public static Color DEFAULT_BACKGROUND_COLOR = Color.white;

  public final static int DRAG_DROP_RANGE_SQUARED = 400;
  public final static int WIDTH_PADING            = 20;
  public final static int HEIGHT_PADING           = 20;

	/* Commands */
  public final static String CUT             = ExtendedTarget.CUT;
  public final static String COPY            = ExtendedTarget.COPY;
  public final static String PASTE           = ExtendedTarget.PASTE;
  public final static String DELETE          = VDT.DELETE;
  public final static String SAVE_TASKS_FILE = VDT.SAVE_TASKS_FILE;


	/** Master root node of the graph */
  protected GraphRoot  rootNode;

	/** Background color of this view */
  protected Color      backgroundColor;

	/** X/Y Offset for drawing so we don't go (negatively) off the page */
  protected int        xOffset, yOffset;

    /** Internal variable used to hold the current receiver of
      * mouseDragged and mouseUp events.
      */ 
  private GraphComponent mouseActionReceiver;

	/** Used to avoid continuously reallocating new Vectors all the time */
  protected Vector     selectedModulesVector;


  public GraphView ( )
  {
    super ( 0, 0, 100, 100 );
    backgroundColor = GraphView.DEFAULT_BACKGROUND_COLOR;
    xOffset = 0;
    yOffset = 0;
    rootNode = new GraphRoot();
    rootNode . setParentView ( this );
    mouseActionReceiver = null;
    selectedModulesVector = new Vector();
    setBuffered ( false );
  }

  public GraphRoot getRootNode ( ) { return rootNode; }

  public Color getBackgroundColor ( ) { return backgroundColor; }
  public void  setBackgroundColor ( Color theBackgroundColor ) 
				      { backgroundColor = theBackgroundColor; }

  public int getXOffset() { return xOffset; }
  public int getYOffset() { return yOffset; }

  public int getSubviewXAdjust()
		    { return getXOffset() + ( GraphView.WIDTH_PADING  / 2 ); }
  public int getSubviewYAdjust()
		    { return getYOffset() + ( GraphView.HEIGHT_PADING / 2 ); }


  public Vector getSelectedModulesVector() { return selectedModulesVector; }


  public void setOffsets ( int newXOffset, int newYOffset )
  {
    int    i;

    for ( i=0;  i < subviews() . count();  i++ )
    {
      ( (View) ( subviews() . elementAt( i ) ) )
	 . moveBy ( newXOffset - getXOffset(),
		    newYOffset - getYOffset() );
    }

    xOffset = newXOffset;
    yOffset = newYOffset;
  }


  public void resizeTo ( Rect theBounds, Rect theActiveRect )
  {
    Rect   originalBounds = bounds();

    int    x = 0,
           y = 0,
           width  = theBounds . width,
           height = theBounds . height;


    if ( theBounds . x < 0 )
    {
      x = - theBounds . x;
    }
    else /* theBounds . x >= 0 */
    {
      width += theBounds . x;
    }

    if ( theBounds . y < 0 )
    {
      y = - theBounds . y;
    }
    else /* theBounds . y >= 0 */
    {
      height += theBounds . y;
    }

    width  += GraphView.WIDTH_PADING;
    height += GraphView.HEIGHT_PADING;

    setBounds ( 0, 0, width, height );
    setMinSize ( width, height );
    setOffsets ( x , y );
    setDirty ( true );

    if ( theActiveRect != null )
    {
      theActiveRect . x += getSubviewXAdjust();
      theActiveRect . y += getSubviewYAdjust();
      scrollRectToVisible ( theActiveRect );
      theActiveRect . x -= getSubviewXAdjust();
      theActiveRect . y -= getSubviewYAdjust();
    }

	/* Try to restore or original scroll-settings...
	 * (And avoid jumping all over the place...)
	 * Also does an implicit:
	 *      ( (ScrollView) superview() ) . updateScrollBars();
	 * that was previously required.
	 */
    if (   ( superview() != null )
	&& ( superview() instanceof Scrollable ) )
    {
      ( (Scrollable) superview() )
	. scrollTo ( originalBounds.x, originalBounds.y );
    }
  }


  public void drawView ( Graphics theGraphics )
  {
	/* lets avoid flashing the screen until we are done... */
    disableDrawing();

    theGraphics . setColor  ( getBackgroundColor() );
	/* Very important -- fillRect with the clipRect rather than 
	 * our bounds() if we are in a scrollview.  Otherwise we hit
	 * a bug in IFC or AWT or wherever.  *sigh*
	 */
    theGraphics . fillRect  ( theGraphics . clipRect() );
    theGraphics . translate ( getSubviewXAdjust(),
			      getSubviewYAdjust() );
    getRootNode() . drawComponents ( theGraphics );
    theGraphics . translate ( - getSubviewXAdjust(),
			      - getSubviewYAdjust() );
    
	/* Lets show what we've done */
    reenableDrawing();
  }


  public boolean isTransparent() { return false; }


  public void ancestorWasAddedToViewHierarchy ( View theAddedView )
  {
    setBounds ( superview() . bounds() );
  }


  public void scrollRectToVisible ( Rect theRect )
  {
    theRect . x += getSubviewXAdjust();
    theRect . y += getSubviewYAdjust();
    super . scrollRectToVisible ( theRect );
    theRect . x -= getSubviewXAdjust();
    theRect . y -= getSubviewYAdjust();
  }


  public void keyDown ( KeyEvent theEvent )
  {
    Scrollable   scrollable;
    int          width, height, i;
    Rect         myBounds;


    if ( theEvent . isTabKey() )
      rootView() . selectViewAfter ( this );


    if ( theEvent . isBackTabKey() )
      rootView() . selectViewBefore ( this );


    if (   ( theEvent . isBackspaceKey() )
	|| ( theEvent . isDeleteKey()    ) )
    {
      performCommand ( GraphView.DELETE, null );
    }


    if (   ( superview() != null               )
	&& ( superview() instanceof Scrollable ) )
    {
      scrollable = (Scrollable) superview();
      myBounds   = bounds();

	/* width/height of our viewable area... */
      width  = superview() . bounds() . width;
      height = superview() . bounds() . height;


      if ( theEvent . isHomeKey() )
	scrollable . scrollTo ( myBounds.x, 0 );

      if ( theEvent . isEndKey() )
	scrollable . scrollTo ( myBounds.x, height - myBounds.height );

      if ( theEvent . isPageUpKey() )
	scrollable . scrollBy ( 0,   (height * 9 / 10 ) );

      if ( theEvent . isPageDownKey() )
	scrollable . scrollBy ( 0, - (height * 9 / 10 ) );

      if ( theEvent . isUpArrowKey() )
	scrollable . scrollBy ( 0,   (height / 10 ) );

      if ( theEvent . isDownArrowKey() )
	scrollable . scrollBy ( 0, - (height / 10 ) );

      if ( theEvent . isLeftArrowKey() )
	scrollable . scrollBy (   ( width / 10 ),  0 );

      if ( theEvent . isRightArrowKey() )
	scrollable . scrollBy ( - ( width / 10 ),  0 );
    }
  }


	/* Forward events taht we want to receive */
  public boolean mouseDown ( MouseEvent theEvent )
  {
	/* Make us the window's latest focused view... */
    setFocusedView();

    if ( getRootNode() != null )
    {
      theEvent . x -= getSubviewXAdjust();
      theEvent . y -= getSubviewYAdjust();
      mouseActionReceiver = getRootNode() . mouseDown ( theEvent );
      theEvent . x += getSubviewXAdjust();
      theEvent . y += getSubviewYAdjust();
      return mouseActionReceiver != null;
    }
    else
    {
      mouseActionReceiver = null;
      return super . mouseDown ( theEvent );
    }
  }

  public void mouseDragged ( MouseEvent theEvent )
  {
    if ( mouseActionReceiver != null )
    {
      theEvent . x -= getSubviewXAdjust();
      theEvent . y -= getSubviewYAdjust();
      mouseActionReceiver . mouseDragged ( theEvent );
      theEvent . x += getSubviewXAdjust();
      theEvent . y += getSubviewYAdjust();
    }
    else
      super . mouseDragged ( theEvent );
  }

  public void mouseUp ( MouseEvent theEvent )
  {
    if ( mouseActionReceiver != null )
    {
      theEvent . x -= getSubviewXAdjust();
      theEvent . y -= getSubviewYAdjust();
      mouseActionReceiver . mouseUp ( theEvent );
      theEvent . x += getSubviewXAdjust();
      theEvent . y += getSubviewYAdjust();
    }
    else
      super . mouseUp ( theEvent );
    mouseActionReceiver = null;
  }

  public void subviewDidResize ( View theSubview )
  {
    getRootNode() . subviewDidResize ( theSubview );
  }


  public DragDestination acceptsDrag ( DragSession theDragSession,
				       int theX, int theY )
  {
    GraphComponent targetComponent;

    theX -= getSubviewXAdjust();
    theY -= getSubviewYAdjust();
    targetComponent = getRootNode() . findNearestComponent ( theX, theY,
					   GraphView.DRAG_DROP_RANGE_SQUARED );
    if ( targetComponent != null )
      return targetComponent . acceptsDrag ( theDragSession, theX, theY );
    else
      return null;
  }


	/* ExtendedTarget Interface */
  public boolean canPerformCommand ( String theCommand ) 
  {
    return (    theCommand . equals ( GraphView.CUT             )
	     || theCommand . equals ( GraphView.COPY            )
	     || theCommand . equals ( GraphView.PASTE           )
	     || theCommand . equals ( GraphView.DELETE          )
	     || theCommand . equals ( GraphView.SAVE_TASKS_FILE ) );
  }

	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
    int             i;

	/* Make us the window's latest focused view... */
    setFocusedView();


    if ( theCommand . equals ( GraphView.CUT ) )
    {
      getSelectedModulesVector() . removeAllElements();

      GraphModule.findSelectedModules ( getRootNode(),
					getSelectedModulesVector() );

      if ( getSelectedModulesVector() . count() > 0 )
      {
	DataComponent.clearCutBuffer();

	for ( i=0;  i < getSelectedModulesVector() . count();  i++ )
	{
	  DataComponent.addToCutBuffer ( 
	     ( (GraphModule) ( getSelectedModulesVector() . elementAt ( i ) ) )
	     . getDataComponent() );

	  ( (GraphModule) ( getSelectedModulesVector() . elementAt ( i ) ) )
	    . removeSelf ( true, true );
	}
      }
    }


    else if ( theCommand . equals ( GraphView.COPY ) )
    {
      getSelectedModulesVector() . removeAllElements();

      GraphModule.findSelectedModules ( getRootNode(),
					getSelectedModulesVector() );

      if ( getSelectedModulesVector() . count() > 0 )
      {
	DataComponent.clearCutBuffer();

	for ( i=0;  i < getSelectedModulesVector() . count();  i++ )
	{
	  DataComponent.addToCutBuffer (
	     ( (GraphModule) ( getSelectedModulesVector() . elementAt ( i ) ) )
	     . getDataComponent() );
	}
      }
    }


    else if ( theCommand . equals ( GraphView.PASTE ) )
    {
      performPasteCommand();
    }


    else if ( theCommand . equals ( GraphView.DELETE ) )
    {
      getSelectedModulesVector() . removeAllElements();

      GraphModule.findSelectedModules ( getRootNode(),
					getSelectedModulesVector() );

      for ( i=0;  i < getSelectedModulesVector() . count();  i++ )
      {
	( (GraphModule) ( getSelectedModulesVector() . elementAt ( i ) ) )
	  . removeSelf ( true, true );
      }
    }


    else if ( theCommand . equals (  GraphView.SAVE_TASKS_FILE ) )
    {
      performSaveTasksFileCommand();
    }

    else
    {
      System.err.println ( "[GraphView:performCommand]  Unknown Command:  \""
			   + theCommand + "\"" );
    }
  }


  void performPasteCommand ()
  {
    GraphComponent  graphComponent;
    int             i;
    boolean         didSucceed = false;


    if ( DataComponent.getCutBufferCount() <= 0 )
    {
      UTIL.runAlert ( "ALERT:",
		      "Paste buffer is empty.",
		      "OK", null, null, rootView() );
      return;
    }

	/* Do we have any selected GraphModules? */
    getSelectedModulesVector() . removeAllElements();
    GraphModule.findSelectedModules ( getRootNode(),
				      getSelectedModulesVector() );
    if ( getSelectedModulesVector() . count() > 0 )
    {
      UTIL.runAlert ( "ALERT:",
		      "Undefined PASTE location:\nToo much is selected.",
		      "OK", null, null, rootView() );
      return;
    }


	/* Do we have exactly one selected GraphComponent? */
    getRootNode() . findSelectedComponents ( getSelectedModulesVector() );
    if ( getSelectedModulesVector() . count() > 1 )
    {
      UTIL.runAlert ( "ALERT:",
		      "Undefined PASTE location:\nToo much is selected...",
		      "OK", null, null, rootView() );
      return;
    }
    if ( getSelectedModulesVector() . count() < 1 )
    {
      UTIL.runAlert ( "ALERT:",
		      "Undefined PASTE location:\nNothing is selected...",
		      "OK", null, null, rootView() );
      return;
    }


	/* Get our selected GraphComponet */
    graphComponent
      = (GraphComponent) ( getSelectedModulesVector() . elementAt ( 0 ) );

    for ( i=0;   i < DataComponent.getCutBufferCount();   i ++ )
    {
      if ( GraphModule . addChildAt ( graphComponent,
				      DataComponent.getCutBufferElement( i ),
				      true, true ) )
      {
	didSucceed = true;
      }
    }
  }


	/* This doesn't really belong here...
	 * But it sure make implemenation easy to put it here...
	 */
  void performSaveTasksFileCommand ( )
  {
    GraphModule    graphModule;

    DataComponent  dataComponent;
    DataFile       dataFile = null;

    for ( int i=0;  i < getRootNode() . getMaxChildren();  i ++ )
    {
      if ( getRootNode() . getChild ( i ) == null )
	continue;

      graphModule = getRootNode() . getChild ( i ) . getOwningGraphModule();

      if ( graphModule == null )
	continue;

      dataComponent = graphModule . getDataComponent();

      while (   (   dataComponent                       != null  )
	     && ( ( dataComponent instanceof DataFile ) == false ) )
      {
	dataComponent = dataComponent . getParent();
      }

      if (   ( dataComponent != null             )
	  && ( dataComponent instanceof DataFile ) )
      {
	if (   ( dataFile != null          )
	    && ( dataFile != dataComponent ) )
	{
	  UTIL.runAlert ( "Error:",
			  "Multiple Data-Files in Diagram.\n"
			  + "Unable to SAVE.\nUse Save-All-Files instead!",
			  "OK", null, null, rootView() );
	  return;
	}
	else
	{
	  dataFile = (DataFile) dataComponent;
	}
      }
    } /* FOR ( 0 <= i < getRootNode() . getMaxChildren() ) */


    if ( dataFile == null )
    {
      UTIL.runAlert ( "Error:",
		      "Unable to locate any Data-Files for this Diagram.\n"
		      + "Unable to SAVE.\nUse Save-All-Files instead!",
		      "OK", null, null, rootView() );
      return;
    }
    else
    {
      DataDestinationFile  dataDestinationFile = new DataDestinationFile();

      if (    dataDestinationFile . setFile ( dataFile . getFilename() )
	   == false )
      {
	UTIL.runAlert (
	      "Error:",
	      "Unable to save file \"" + dataFile . getFilename() + "\".",
	      "OK", null, null, rootView() );
	return;
      }
      else
      {
	dataFile . generate ( dataDestinationFile,
			      DataComponent.ENTIRE_OBJECT );
	dataDestinationFile . close();
      }
    }
  }

}

