
import netscape.application.*;
import netscape.util.*;
import netscape.constructor.*;

public class OpenTaskWindowHandler extends FileTaskWindowHandler
{
	/* Constant (Special) Filenames */
  public final static String OPEN_TASK_WINDOW_PLAN_NAME
						     = "OpenTaskWindow.planb";

	/* TASK_WINDOW_PLAN Objects */
  public final static String INTERNAL_WINDOW_BUTTON  = "InternalWindowButton";
  public final static String EXTERNAL_WINDOW_BUTTON  = "ExternalWindowButton";
  public final static String CANCEL_BUTTON           = "CancelButton";
  public final static String OPEN_BUTTON             = "OpenButton";

	/* Commands */
  public final static String CANCEL_COMMAND          = "Cancel";
  public final static String OPEN_COMMAND            = "Open";

	/* Constants */
  public final static String   TITLE                 = "Open Task";


	/* Class variables */
  protected static Plan                   openTaskWindowPlan    = null;
  protected static OpenTaskWindowHandler  openTaskWindowHandler = null;

	/* Class Methods */
  protected static synchronized Plan getOpenTaskWindowPlan()
  {
    if ( OpenTaskWindowHandler.openTaskWindowPlan == null )
    {
      OpenTaskWindowHandler.openTaskWindowPlan
	= WindowHandlerAbstractBaseClass.getPlan (
		OpenTaskWindowHandler.OPEN_TASK_WINDOW_PLAN_NAME,
		"[OpenTaskWindowHandler:getOpenTaskWindowPlan]  "
		+ "Error:  Failure - Unable to access "
		+ "OPEN_TASK_WINDOW_PLAN_NAME" );
    }
    return OpenTaskWindowHandler.openTaskWindowPlan;
  }

  public static OpenTaskWindowHandler getOpenTaskWindowHandler()
  {
    if ( OpenTaskWindowHandler.openTaskWindowHandler == null )
    {
      OpenTaskWindowHandler.openTaskWindowHandler
	= new OpenTaskWindowHandler();
    }
    return OpenTaskWindowHandler.openTaskWindowHandler;
  }



	/* Instance Variables */
  protected Button  internalWindowButton,
		    externalWindowButton,
		    cancelButton,
		    openButton;


	/* Constructors */
  public OpenTaskWindowHandler ( )
			{ createAndInitialize ( false ); }

  public OpenTaskWindowHandler ( boolean  theUseExternalWindow )
			{ createAndInitialize ( theUseExternalWindow ); }

	/* Constructors can't be synchronized. */
	/* And this needs to be synchronized since we use the static plans */
  private synchronized
      boolean createAndInitialize ( boolean theUseExternalWindow )
  {
	/* Let the super-class do it's thing... */
    if ( super . createAndInitialize (
			         OpenTaskWindowHandler.getOpenTaskWindowPlan(),
				 null,
				 OpenTaskWindowHandler.OPEN_COMMAND,
				 theUseExternalWindow,
				 true,
				 OpenTaskWindowHandler.TITLE,
				 "OpenTaskWindowHandler",
				 null )
	 == false )
    {
      return false;
    }

	/* Lets find our instance variables */
    internalWindowButton
      = (Button)
         OpenTaskWindowHandler.getOpenTaskWindowPlan().componentNamed (
			      OpenTaskWindowHandler . INTERNAL_WINDOW_BUTTON );
    externalWindowButton
      = (Button)
         OpenTaskWindowHandler.getOpenTaskWindowPlan().componentNamed (
			      OpenTaskWindowHandler . EXTERNAL_WINDOW_BUTTON );
    cancelButton
      = (Button)
         OpenTaskWindowHandler.getOpenTaskWindowPlan().componentNamed (
				       OpenTaskWindowHandler . CANCEL_BUTTON );
    openButton
      = (Button)
         OpenTaskWindowHandler.getOpenTaskWindowPlan().componentNamed (
				         OpenTaskWindowHandler . OPEN_BUTTON );

	/* Error-check */
    if (   ( internalWindowButton == null )
	|| ( externalWindowButton == null )
	|| ( cancelButton         == null )
	|| ( openButton           == null ) )
    {
      System.err.println ( "[OpenTaskWindowHandler:createAndInitialize]  "
	   + "Warning:  One or more Plan objects is NULL!!!\n"
	   + "    internalWindowButton...: " + internalWindowButton + "\n"
	   + "    externalWindowButton...: " + externalWindowButton + "\n"
	   + "    cancelButton...........: " + cancelButton         + "\n"
	   + "    openButton.............: " + openButton           + "\n"
	   + "------------------------------------------------------------" );
    }


	/* Reset the plan so that we can use it again later on... */
    OpenTaskWindowHandler.getOpenTaskWindowPlan() . releaseObjects();

	/* Reconfigure our objects... */


	/* Report that we succeeded. */
    return true;
  }



  public Button      getInternalWindowButton() { return internalWindowButton; }
  public Button      getExternalWindowButton() { return externalWindowButton; }
  public Button      getCancelButton()         { return cancelButton;         }
  public Button      getOpenButton()           { return openButton;           }


	/* ExtendedTarget Interface */
  public boolean canPerformCommand ( String theCommand ) 
  {
    return 
      (    theCommand . equals ( OpenTaskWindowHandler.CANCEL_COMMAND   )
	|| theCommand . equals ( OpenTaskWindowHandler.OPEN_COMMAND     )
	|| super . canPerformCommand ( theCommand ) );
  }


	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
    String  openWindowCommand = null;
    Target  target;

	/* Handle canceling... */
    if ( theCommand . equals( OpenTaskWindowHandler.CANCEL_COMMAND ) )
    {
      getWindow() . hide();
    }

	/* Handle successful conclusion. */
    else if ( theCommand . equals ( OpenTaskWindowHandler.OPEN_COMMAND ) )
    {
      if (   ( getTaskListView() . selectedItem() != null )
	  && ( getTaskListView() . count()         > 0    ) )
      {
	    /* Create a new window...  */
	if ( getInternalWindowButton() . state() )
	  openWindowCommand = VDT.NEW_INTERNAL_WINDOW;
	else if ( getExternalWindowButton() . state() )
	  openWindowCommand = VDT.NEW_EXTERNAL_WINDOW;
	else
	{
	  System.err.println ( "[OpenTaskWindowHandler:performCommand(OPEN)]  "
	       + "Warning:  Neither Internal or External window buttons are "
	       + "selected.  Assuming Internal Window..." );
	  openWindowCommand = VDT.NEW_INTERNAL_WINDOW;
	}

	target = TargetChain . applicationChain()
			     . targetForCommand ( openWindowCommand );

	if ( target != null )
	{
	  Application . application()
		      . performCommandLater ( target, openWindowCommand,
					      getTaskListView()
					        . selectedItem() . data() );
	}
	else
	{
	  System.err.println ( "[OpenTaskWindowHandler:performCommand(OPEN)]  "
		 + "Error: Unable to find TARGET for Opening New Window. (\""
		 + openWindowCommand + "\")" );
	}

	    /* Close the open-task window... */
	getWindow() . hide();
      }
      else
      {
	runAlert ( "Error:", "No Task has been Selected.",
		   "OK", null, null );
      }
    } /* else if ( theCommand . equals ( OPEN_COMMAND ) ) */


	/* Otherwise, let the superclass deal with it... */
    else
    {
      super . performCommand ( theCommand, theData );
    }

  } /* public void performCommand ( String theCommand, Object theData ) */

}
