
/*
 * This is the place for code that is utilized by compiled TDL code.
 * Naturally, this code can be used directly.  But its purpose is to
 * support (implement) specific TDL features.
 */

#ifndef TDL_H
#define TDL_H

 /* Standard C++ Header Files */
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <strstream.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <sys/wait.h>

    /* TCM defines a const variable called MAXINT.
     * Unfortunately, so does math.h.  Only math.h #defines MAXINT...
     */
#undef MAXINT


 /* TCM Header Files */
#include <tcm.h>


 /* TDL Header Files */
#include "_TDL_BaseAction.H"
#include "stdtypes.H"
#include "tdldef.H"
#include "_TDL_InterfaceToTCM.H"
#include "_TDL_ActionOrVoid.H"
#include "_TDL_AllocationFunction.H"


/* Support for the return-value from:                         *
 *  extern  _TDL_Initialize_Return_Class TDL_Initialize(...); */
class _TDL_Initialize_Return_Class
{
private: char foo;  /* Classes with no associated data bother some compilers */
public:  _TDL_Initialize_Return_Class() : foo ( char(NULL) ) {}
         ~_TDL_Initialize_Return_Class() {} /* NOT Virtual */
         _TDL_Initialize_Return_Class connectAgent (
			    const char * theAgent,
			    const char * theCentralHost = (const char *)NULL );
};

/* We wish to replace TCM_Initialize() with TDL_Initialize(), which will be
 * invoked automatically, and will return an instance of a class which can be
 * used to connect distributed agents. Class/Functions are defined in tdldef.C.
 */

extern _TDL_Initialize_Return_Class TDL_Initialize (
		      /*BOOLEAN*/int theHideTcmMessages,
		      /*BOOLEAN*/int theHideTdlMessages  );
extern _TDL_Initialize_Return_Class TDL_Initialize (
		      const char *   theAgent,
		      const char *   theCentralHost,
		      /*BOOLEAN*/int theHideTcmMessages,
		      /*BOOLEAN*/int theHideTdlMessages );

	/* Convenience versions... */
inline _TDL_Initialize_Return_Class TDL_Initialize (
		      /*BOOLEAN*/int theHideTcmMessages,
		      /*BOOLEAN*/int theHideTdlMessages,
		      const char *   theAgent,
		      const char *   theCentralHost     = (const char *) NULL )
	  { return TDL_Initialize ( theAgent, theCentralHost,
				    theHideTcmMessages, theHideTdlMessages ); }

inline _TDL_Initialize_Return_Class TDL_Initialize (
		      /*BOOLEAN*/int theHideMessages    = TRUE )
	  { return TDL_Initialize ( theHideMessages, theHideMessages ); }

inline _TDL_Initialize_Return_Class TDL_Initialize (
		      /*BOOLEAN*/int      theHideMessages,
		      const char *   theAgent,
		      const char *   theCentralHost     = (const char *) NULL )
	  { return TDL_Initialize ( theAgent, theCentralHost,
				    theHideMessages, theHideMessages ); }

inline _TDL_Initialize_Return_Class TDL_Initialize (
		      const char *   theAgent,
		      const char *   theCentralHost     = (const char *) NULL,
		      /*BOOLEAN*/int theHideMessages    = TRUE )
	  { return TDL_Initialize ( theAgent, theCentralHost,
				    theHideMessages, theHideMessages ); }

inline _TDL_Initialize_Return_Class TDL_Initialize (
		      const char *   theAgent,
		      /*BOOLEAN*/int theHideMessages )
	  { return TDL_Initialize ( theAgent, (const char *)NULL,
				    theHideMessages, theHideMessages ); }

inline _TDL_Initialize_Return_Class TDL_Initialize (
		      const char *   theAgent,
		      /*BOOLEAN*/int theHideTcmMessages,
		      /*BOOLEAN*/int theHideTdlMessages )
	  { return TDL_Initialize ( theAgent, (const char *)NULL,
				    theHideTcmMessages, theHideTdlMessages ); }







	/* Create this as a global, so that spawns outside of tasks can
	 * refer to "PARENT".  (Ie: For the #define to work right...)
	 * Now used in place of TCM_RootNode() to provide a rather hideous
	 * mechanism for inserting external nodes (spawns outside of tasks)
	 * into the TCM Task Tree at locations other than the TCM_RootNode.
	 */
extern TCM_Task_Tree_Ref const & _TDL_ENCLOSING_TASK; /* = TCM_RootNode() */

	/* Macro to mark variables as "USED" to the compiler, so as to   *
	 * prevent variable-not-used warnings on TDLC-created variables. *
	 * TDLC utilizes (generates) _TDL_MARKUSED(...) commands.        *
	 * If users utilize MARKUSED(...), the two can be distinguished. */
#define _TDL_MARKUSED(X)  ((void)(&(X)))
#define MARKUSED(X)       ((void)(&(X)))

	/* Monitor functions need to reference _TDL_ENCLOSING_TASK */
#define getNumberOfTriggers()    ( TCM_NumTriggers    ( _TDL_ENCLOSING_TASK ) )
#define getNumberOfActivates()   ( TCM_NumActivations ( _TDL_ENCLOSING_TASK ) )

	/* Note:  This is a case-insenstive macro.  The TDLC program         *
	 * automatically translates the keyword "TRIGGER" to all upper-case. */
#define TRIGGER()                ( TCM_Trigger ( _TDL_ENCLOSING_TASK ) )


	/* Trival means for user to access TDL Tasks as             *
	 * "const TCM_Task_Tree_Ref &"'s for use with TCM functions */
#define TDL_REF(X)   _TDL_SpawnedTasks [ # X ]


	/* Folks may want to implement an exception handler for BaseException*/
#define TDL_BaseException _TDL_BaseException

	/* Sometimes it's nice to know the Derived-Exception-name of */
	/* a Base-Exception.  Only useful inside Exception Handlers. */
#define TDL_DerivedExceptionName(X)  ((X).getExceptionName())

	/* Somewhat complicated way to access arguments from Tasks while
	 * outside the scope of those tasks...  Useful for POSTPONE'ed tasks.
	 * Ie:  TDL_TaskData_Ref          taskDataRef;
	 *      const TCM_Task_Tree_Ref & task;
         *      Goal foo ( int i ) { taskDataRef = TDL_TASKDATA();
         *                           task        = TDL_REF ( PARENT );
         *                           POSTPONE; }
         *      void bar() { cerr << _TDL_ARG(taskDataRef, foo, i);
         *                   TDL_SUCCESS ( task ); }
	 */
typedef TCM_Action_Ref TDL_TaskData_Ref;
typedef _Action        TDL_TaskData;
#define TDL_TASKDATA()    (this)
#define TDL_ARG( theTaskDataRef, theTaskName, theArgumentName )            \
 ( ((_TDL_ ## theTaskName *)((theTaskDataRef).operator*())) -> theArgumentName)
#define TDL_TASK_ARG( theTaskRef, theTaskName, theArgumentName )           \
 ( ((_TDL_ ## theTaskName *)(TCM_GetActualAction(theTaskRef).operator*())) \
   -> theArgumentName)

	/* POSTPONE'ed Tasks need a means to be marked SUCCEEDED or FAILED. */

	/* Ie: void bar() { TDL_SUCCESS ( task ); } */
#define TDL_SUCCESS( theTaskRefToMarkAsSucceeded )                            \
			   _TDL_DO_TCM_SUCCESS ( theTaskRefToMarkAsSucceeded )

	/* Note:  TheFailureArgs must be surrounded by parenthesis!!!!    */
	/* Ie: void bar() { TDL_FAIL ( task, engineOnFire, ("f3", 1) ); } */
#define TDL_FAIL( theTaskRefToFail, theFailureName, theFailureArgs )          \
  _TDL_DO_TCM_FAIL ( (theTaskRefToFail),                                      \
		     _TDL_Exception_ ## theFailureName ## ::exceptionMatches, \
		     _TDL_CreateException_ ## theFailureName theFailureArgs  )

	/* And POSTPONE'd Exception-Handlers need to be */ 
	/* able to access their Exception's data.       */
#define TDL_FAIL_DATA_ARG( theTaskDataRef, theExceptionName, theArgumentName )\
			    ( ( (_TDL_Exception_ ## theExceptionName *)       \
				(TCM_FailureDataRefCount ( theTaskDataRef ) ) \
			       ) -> theArgumentName )



	/* Macros to test if a TASK has completed execution. */
#define TDL_TASK_IS_COMPLETED( theTCMTaskTreeRef )                            \
	 ( TCM_IsPast ( TCM_EndOf ( TCM_AchievingOf ( theTCMTaskTreeRef ) ) ) )
#define TDL_IS_COMPLETED( theTask )                                           \
	 ( TDL_TASK_IS_COMPLETED ( TDL_REF ( theTask ) ) )

	/* Macros to test if a TASK has started. */
#define TDL_TASK_IS_STARTED( theTCMTaskTreeRef )                              \
	 ( TCM_IsPast ( TCM_StartOf ( TCM_HandlingOf ( theTCMTaskTreeRef ) ) ))
#define TDL_IS_STARTED( theTask )                                             \
	 ( TDL_TASK_IS_STARTED ( TDL_REF ( theTask ) ) )

	/* Macros to test if a TASK is of a particular Task-type. *
	 *   Returns TRUE for direct instances [SPAWN foo();],    *
	 *   both Monitor and Monitor-child instances ["ACT-"],   *
	 *   and SPAWN-with-WAIT macro instances ["-auto,wait"].  */
#define TDL_TASK_INSTANCE_OF( theTCMTaskTreeRef, theTaskName )                \
	 ( TDL::checkForTaskInstanceOf ( TCM_NodeName ( theTCMTaskTreeRef ),  \
					 # theTaskName ) )
#define TDL_INSTANCE_OF( theTask, theTaskName )                               \
	 ( TDL_TASK_INSTANCE_OF ( TDL_REF ( theTask ), theTaskName ) )


	/* Macro to implement the real RESUME functionality. *
	 * Used internally to TDL.                           */
#define _TDL_INTERNAL_RESUME( theTaskClassName, theTCMTaskTreeRefToResume )   \
    ( (theTaskClassName *)                                                    \
      ( ( (_TDL_BaseAction *)                                                 \
	  ( TCM_GetActualAction ( theTCMTaskTreeRefToResume ) . operator*() ) \
	) -> setTcmTaskTreeNode ( theTCMTaskTreeRefToResume . operator *() )  \
      )                                                                       \
    )  -> resume

#define TDL_RESUME_ARGUMENT_DECLARATIONS  \
	  const TCM_Task_Tree_Ref & _TDL_ENCLOSING_TASK


	/* KLUDGE to allow us to spawn children under an alternate
	 * Task Tree Node inside a task.   Untested.  Use at own risk.
	 */
#define TDL_BlackMagic_SwitchParentTaskNodeTo( theTCMTaskTreeRef )            \
	   _TDL_SpawnedTasks . setEnclosingTaskReference ( theTCMTaskTreeRef )



	/* Originally intended to simplify testing of Distributed code,
	 * these functions may prove useful for developing Distributed
	 * applications as well.
	 */
extern void TDL_ForkChildren (
	      void ( *     theFunctions [] )(u_int4),
	      const char * theNames [],
	      u_int4       theNumberOfFunctions,
	      MSecs        theDurationTimeInMilliSeconds,
	      MSecs        theTimeBetweenStartingProcessesInMilliseconds = 0,
	      BOOLEAN      thePrintTimeStampsForDebugging = FALSE            );

	/* Convenience function to simplify  Central/Server/Client  case. */
extern void TDL_ForkChildren (
		  void ( * theServerFunction  )(u_int4),
		  void ( * theClientFbunction )(u_int4),
		  MSecs    theDurationTimeInMilliSeconds,
		  MSecs    theTimeBetweenStartingProcessesInMilliseconds = 0,
		  BOOLEAN  thePrintTimeStampsForDebugging = FALSE            );





#ifndef _TDL_INTERNAL_

#include "_TDL_Snode.H"
#include "_TDL_Slist.H"
#include "_TDL_Dnode.H"
#include "_TDL_Dlist.H"
#include "_TDL_ArrayList.H"
#include "_TDL_NamesList.H"
#include "_TDL_TreeNode.H"
#include "_TDL_TreeNodeBranch.H"
#include "_TDL_IterationIndex.H"
#include "_TDL_TDLStatement.H"
#include "_TDL_SpawnStatement.H"
#include "_TDL_SpawnStatementData.H"
#include "_TDL_SpawnStatementTreeNode.H"
#include "_TDL_Exceptions.H"
#include "_TDL_Constraints.H"
#include "_TDL_WithStatement.H"
#include "_TDL_WithStatementData.H"
#include "_TDL_WithStatementTreeNode.H"
#include "_TDL_HandleManager.H"
#include "_TDL_functions_2.H"
#include "_TDL_DistributedRegistryEntry.H"

#endif /* ifndef _TDL_INTERNAL_ */



     /* Miscellaneous:
      *     Does our compiler conform to the new (1998) ANSI C++ Standard?
      *
      * Note:  The new [1998] ANSI C++ Standard has changed a number of things.
      * At present, both g++ 2.8.1 and egcs support the "new" Standard.  Which,
      * unfortunately, is syntactically incompatible with the "old" Standard
      * supported by g++ 2.7.2.x.
      */
#if    ( ! defined ( ANSI_STANDARD_CXX )          ) \
    && ( __GNUC__ >= 2 && ( __GNUC_MINOR__ >= 8 ) )

#define ANSI_STANDARD_CXX 1

#endif 


#endif /* TDL_H */

