
/*
 * This class is utilized with respect to Distributed Tasks.
 * It allows task's to register themselves via the initialization a
 * static variable containing an instance of this class.
 * Later on, in TDL_INITIALIZE(), we exploit these registrations to
 * register these distributed-tasks appropriately with TCM.
 */

#define _TDL_INTERNAL_
#include "tdl.H"
#include "_TDL_Snode.H"
#include "_TDL_Slist.H"

#ifndef _TDL_DISTRIBUTED_REGISTRY_ENTRY
#define _TDL_DISTRIBUTED_REGISTRY_ENTRY

typedef TCM_Task_Tree_Ref ( * _TDL_DistributedAllocationFunctionType ) 
				  ( const char *,
				    const char *,
				    BOOLEAN       );

typedef TCM_Action_Ref ( * _TDL_DistributedCreateRemoteActionFunctionType ) 
				  ( const TCM_Task_Tree_Ref &,
				    void *,
				    void *  );

class _TDL_DistributedRegistryEntry : _TDL_Snode
{
	/* Class methods */
protected:
  static _TDL_Slist & getAllDistributedRegistriesList();

  static status_t addRegistryToMasterList (
		 _TDL_DistributedRegistryEntry * theDistributedRegistryEntry )
    { return _TDL_DistributedRegistryEntry::getAllDistributedRegistriesList()
	       . appendNode ( theDistributedRegistryEntry ); }

public:
  static int4              registerAllDistributedRegistryEntries();
  static TCM_Task_Tree_Ref genericDistributedRemoteTaskAlloctionFunction ( 
						     void * theTcmClientData );



	/* Instance variables */
protected:
  const char *                           taskName;
  const char *                           distributedRemoteTaskName;

  _TDL_DistributedAllocationFunctionType distributedAllocationFunction;

  _TDL_DistributedCreateRemoteActionFunctionType 
					 distributedCreateRemoteActionFunction;

  const char *                           distributedFormatString;
  BOOLEAN                                hasBeenRegisteredWithTcm;



	/* Instance methods */
public:
  _TDL_DistributedRegistryEntry (
    const char *                      theTaskName,
    const char *                      theDistributedRemoteTaskName,
    _TDL_DistributedAllocationFunctionType
				      theDistributedAllocationFunction,
    _TDL_DistributedCreateRemoteActionFunctionType
				      theDistributedCreateRemoteActionFunction,
    const char *                      theDistributedFormatString )

    : _TDL_Snode                    (                                        ),
      taskName                      ( theTaskName                            ),
      distributedRemoteTaskName     ( theDistributedRemoteTaskName           ),
      distributedAllocationFunction ( theDistributedAllocationFunction       ),
      distributedCreateRemoteActionFunction (
				    theDistributedCreateRemoteActionFunction ),
      distributedFormatString       ( theDistributedFormatString             ),
      hasBeenRegisteredWithTcm      ( FALSE                                  )
    { _TDL_DistributedRegistryEntry::addRegistryToMasterList ( this ); }


  virtual ~_TDL_DistributedRegistryEntry();

  
  const char * getTaskName() const
    { return taskName; }

  const char * getDistributedRemoteTaskName() const
    { return distributedRemoteTaskName; }

  _TDL_DistributedAllocationFunctionType
  getDistributedAllocationFunction()     const
    { return distributedAllocationFunction; }

  _TDL_DistributedCreateRemoteActionFunctionType 
  getDistributedCreateRemoteActionFunction()     const
    { return distributedCreateRemoteActionFunction; }

  const char * getDistributedFormatString() const
    { return distributedFormatString; }

  BOOLEAN  getHasBeenRegisteredWithTcm() const
    { return hasBeenRegisteredWithTcm; }


	/* Abstract iostreamBase-interface method */
  virtual ostream & printObject ( ostream    & theOstream,
				  const char * theIndentString = "") const;

protected:
  void     setHasBeenRegisteredWithTcm()
    { hasBeenRegisteredWithTcm = TRUE; }


	/* These operations are disabled (prohibited) & unimplemented */
private:
  _TDL_DistributedRegistryEntry (
		       const _TDL_DistributedRegistryEntry & theObjectToCopy );
  _TDL_DistributedRegistryEntry & operator = (
		       const _TDL_DistributedRegistryEntry & theObjectToCopy );
};

#endif /* _TDL_DISTRIBUTED_REGISTRY_ENTRY */
